# -*- coding: utf-8 -*-
#
#    GlyphAtlas Character Selector and Character Map
#    Copyright (C) 2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import binascii
from collections import OrderedDict, ChainMap, namedtuple
from collections.abc import Hashable
from functools import partial

from typeatlas.compat import QtCore, QtGui, QtWidgets, QtModelProxies
from typeatlas.compat import Qt, Slot, Signal, setResizeMode
from typeatlas.compat import configuredStyle, usesPySide

from typeatlas.guicommon import FontRenderingChoice
from typeatlas.charinfo import CharacterDatabase, UNICODE_DATA_FIELDS
from typeatlas.langutil import LanguageDatabase, install_locales
from typeatlas.datastore import MetadataCache
from typeatlas.options import Options

from typeatlas.uitools import QtExecutor
from typeatlas.uitools import setDefaultFontFamilies
from typeatlas.uitools import getIcon, getImage

from typeatlas.util import OrderedSet, sorteddict, bytedata_to_int, debugmsg
from typeatlas.util import disable_debug, MaybeIterable, generic_type
from typeatlas.langutil import _, N_, H_, textlang
from html import escape as htesc

from typeatlas import compat, proginfo
from typeatlas import fontlist, fontgrid, qfontlist
from typeatlas import charsets
import typeatlas
try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False
else:
    if TYPE_CHECKING:
        import typeatlas.gui

Union = generic_type('Union')
Optional = generic_type('Optional')


LibraryLike = 'Union[GlyphAtlasLibrary, typeatlas.gui.TypeAtlasLibrary]'


class GlyphAtlasLibrary(QtWidgets.QWidget):

    """Manager of main GlyphAtlas windows, shared attributes between
    those instances. A reduced version of TypeAtlasLibrary, which can
    be used in its place"""

    instance = None

    def __init__(self, splash: QtWidgets.QSplashScreen=None,
                       atlasLibrary: 'typeatlas.gui.TypeAtlasLibrary'=None,
                       *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.options = Options.getInstance()
        if atlasLibrary is not None:
            self.options.load()

        self.metadataCache = MetadataCache.getInstance()

        self.charDb = CharacterDatabase.getInstance()
        self.langDb = LanguageDatabase.getInstance()
        self.fontDb = QtGui.QFontDatabase()

        #if splash is not None:
        #    splash.showMessage(_('Loading cache...'), splash.msgAlignment)
        #debugmsg("Loading cache...")
        #
        #self.metadataCache.load()

        #if splash is not None:
        #    splash.showMessage(_('Loading font list...'), splash.msgAlignment)
        #debugmsg("Loading font list...")
        #self.executor = executor = QtExecutor(self.options.executablePaths,
        #                                      parent=self)
        #self.finder = qfontlist.QtFontFinder(self.fontDb, executor=executor,
        #                                     metadata_cache=self.metadataCache)
        #self.fontFamilies = list(self.finder.families())
        #setDefaultFontFamilies(self.fontFamilies)

        #debugmsg("Saving metadata cache...")
        #self.metadataCache.autosave()

        self.executor = executor = QtExecutor(self.options.executablePaths,
                                              parent=self)
        self.finder = qfontlist.QtFontFinder(self.fontDb, executor=executor)

        if splash is not None:
            splash.showMessage(_('Populating unicode info...'),
                               splash.msgAlignment)
        debugmsg("Populating unicode info...")
        self.charDb.add_registries()
        self.charDb.populate(download=False, deep=True)
        if splash is not None:
            splash.showMessage(_('Populating language info...'),
                               splash.msgAlignment)
        debugmsg("Populating language info...")
        self.langDb.populate()

        debugmsg("DONE.")

        self.openWindows = set()

    @classmethod
    def getInstance(cls, *args, fallback: bool=True,
                                create: bool=True, **kwargs) -> LibraryLike:
        """Get the singleton instance of the class. If create=False is
        passed, it is not created if it does not exist and None is
        returned.

        By default, we attempt to get the TypeAtlas library instead.
        If that is not OK, pass fallback=False.
        """
        if cls.instance is None:

            if fallback:
                from typeatlas.gui import TypeAtlasLibrary
                instance = TypeAtlasLibrary.getInstance(create=False)
                if instance is not None:
                    return instance

            if not create:
                return None

            cls.instance = cls(*args, **kwargs)
        return cls.instance

    @Slot()
    def quit(self):
        """Quit GlyphAtlas."""
        sys.exit(0)

    @Slot()
    def newGlyphAtlasWindow(self):
        """Open a new main GlyphAtlas window."""
        window = GlyphAtlas(self)
        window.closed.connect(self._windowClosed)
        window.show()
        self.openWindows.add(window)
        return window

    @Slot()
    def _windowClosed(self):
        """A window was close it. Clear it, prepare its state for
        saving, and save and quit if needed."""
        window = self.sender()
        self.openWindows.discard(window)
        window.deleteLater()
        if not self.openWindows:
            self.quit()


class GlyphSelectorMixin():

    """A mixin used for our character map main windows, both of a standalone
    GlyphAtlas, and of a character selector dialog - CharSelectDialog."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.charsetSubmenus = {}

        self.encoding = None
        self.interpretation = None
        self.charsetCodeInfo = None

        self.selectFromFont = False

        self._interpretationActions = []

    def getCharsetSubmenu(self, key: Hashable, name: str) -> QtWidgets.QMenu:
        """Get a character set submenu."""
        if key in self.charsetSubmenus:
            return self.charsetSubmenus[key]

        result = self.charsetMenu.addMenu(name)
        self.charsetSubmenus[key] = result
        return result

    def populateCharsetMenu(self):
        """Populate the character sets menu."""
        charsetMenu = self.charsetMenu

        for name, encoding in ChainMap(charsets.morse_codes,
                                       charsets.encodings).items():

            if encoding.morse:
                submenu = self.getCharsetSubmenu('morse', _('Telegraphy'))
            else:
                submenu = self.getCharsetSubmenu(encoding.category(),
                                                 encoding.category_label())
            action = submenu.addAction(encoding.label())
            iconName = encoding.get_icon()
            if iconName is not None:
                action.setIcon(getIcon(iconName))
            action.setData(encoding)
            action.triggered.connect(self._charsetClicked)

        action = charsetMenu.addAction(_('Unicode'))
        action.setData(None)
        action.triggered.connect(self._charsetClicked)

    def populateInterpretationMenu(self):
        """Populate the interpretations menu."""
        menu = self.interpretationMenu

        for group in charsets.InterpretationGroup.available:
            actionGroup = QtWidgets.QActionGroup(self)
            if group.exclusive:
                actionGroup.setExclusive(True)

            submenu = menu.addMenu(_(group.label))
            if group.icon is not None:
                submenu.setIcon(getIcon(group.icon))

            for interpretation in group.interpretations:
                action = submenu.addAction(_(interpretation.label))
                if interpretation.icon is not None:
                    action.setIcon(getIcon(interpretation.icon))
                action.setData(interpretation)
                action.triggered.connect(self._interpretationClicked)

                action.setCheckable(True)
                if interpretation.default:
                    action.setChecked(True)
                actionGroup.addAction(action)

                self._interpretationActions.append(action)

        self._interpretationClicked()

    def _charsetClicked(self):
        """A new character set was selected, switch to it."""
        action = self.sender()
        #self.encoding = charsets.get_encoding(charset)
        self.encoding = action.data()
        if self.interpretation and self.encoding is not None:
            self.encoding = self.encoding.interpretation(self.interpretation)
        self.updateFontGrid()

    def _interpretationClicked(self):
        """A new interpretation set was selected, switch to it."""
        selected = []
        for action in self._interpretationActions:
            if action.isChecked():
                selected.append(action.data())

        self.interpretation = charsets.InterpretationSet(selected)

        if self.encoding is None:
            return

        self.encoding = self.encoding.interpretation(self.interpretation)
        self.updateFontGrid()

    def updateFontGrid(self):
        """Update the font grid character map widget."""
        encoding = self.encoding
        charDb = self.atlasLibrary.charDb

        if encoding is None or encoding.unicode:
            if self.selectFromFont:
                self.fontGrid.setFontItem(self.fontItem)
            else:
                self.fontGrid.setFontItem(self.fontItem,
                                          charBlocks=charDb.script_blocks)
            self.charsetCodeInfo = None
            return

        if encoding.morse:
            characters = OrderedDict((alpha,
                                      fontgrid.CharSymbolInfo(
                                          alpha,
                                          display=morse,
                                          encoding=encoding))
                                     for alpha, morse
                                        in encoding.codepoint_telegraph_codes().items())

            self.charsetCodeInfo = characters
            self.fontGrid.setFontItem(self.fontItem, characters=characters)
            return

        codepointBytes = encoding.codepoint_bytes(charDb)
        codepointBytes = sorteddict(codepointBytes, valuekey=bytedata_to_int)

        characters = OrderedDict((code,
                                  fontgrid.CharSymbolInfo(code, bytedata,
                                                          encoding=encoding))
                                 for code, bytedata in codepointBytes.items())

        self.charsetCodeInfo = characters
        self.fontGrid.setFontItem(self.fontItem, characters=characters)

    @Slot()
    def selectGenericSymbol(self):
        """Switch the character selector to generic symbol (non-font specific).
        For example, letters, regular emojis.

        This tends to display all characters from the character set, or from
        unicode.
        """
        self.selectFromFont = False
        self.updateFontGrid()

    @Slot()
    def selectSymbolFromFont(self):
        """Switch the character selector to a font-specific symbol, likely from
        a Unicode private use area. The user is asked for a font with a dialog.

        This tends to display only the characters from the font.
        """
        return self._chooseFont(True)

    @Slot()
    def selectDisplayFont(self):
        """Switch the display font in general symbol mode after asking the
        user. This opens a dialog."""
        return self._chooseFont(False)

    def _chooseFont(self, selectFromFont: bool=False, force: bool=False):
        """Change the font, asking the user in a dialog.

        If selectFromFont=True is passed, it means we're switching to
        symbol-from-font mode (likely PUA symbol), whereas if it is
        False, we simply change the default display font, and still
        select a general symbol from anywhere.

        If force=True, this allows us to change the display font if
        we're in symbol-from-font mode.
        """
        item = self.fontItem

        if selectFromFont:
            title = _("Select symbol from a specific font")
        else:
            if self.selectFromFont and not force:
                QtWidgets.QMessageBox.critical(
                    self, _("Cannot change display font"),
                          _("You are selecting a symbol from a font, "
                            "the display font cannot be changed"))
                return

            title = _("Choose display font")

        initial = self.atlasLibrary.fontDb.font(item.family, item.style,
                                                self.renderingChoice.gridSize)
        res = QtWidgets.QFontDialog.getFont(initial)

        if usesPySide:
            ok, qfont = res
        else:
            qfont, ok = res

        if ok:
            self.fontItem = self.atlasLibrary.finder.substitute(
                                    qfont.family(), qfont.styleName())
            self.selectFromFont = selectFromFont
            self.updateFontGrid()


class GlyphAtlas(GlyphSelectorMixin, QtWidgets.QMainWindow):

    """GlyphAtlas main window."""

    def __init__(self, atlasLibrary: 'LibraryLike'=None,
                       parent: QtWidgets.QWidget=None,
                       splash: QtWidgets.QSplashScreen=None):

        super().__init__(parent)

        if atlasLibrary is None:
            atlasLibrary = GlyphAtlasLibrary.getInstance(splash)

        self.resize(atlasLibrary.options.glyphAtlasSize)

        self.setWindowTitle(_("GlyphAtlas character map"))
        self.setWindowIcon(getIcon('glyphatlas'))

        mainMenu = self.menuBar()
        self.charsetMenu = mainMenu.addMenu(_("&Character set"))
        self.interpretationMenu = mainMenu.addMenu(_('&Alternative characters'))

        self.populateCharsetMenu()
        self.populateInterpretationMenu()

        self.atlasLibrary = atlasLibrary
        self.renderingChoice = FontRenderingChoice()

        self.fontGrid = fontgrid.FontGrid(atlasLibrary.charDb,
                                          fontDb=atlasLibrary.fontDb,
                                          langDb=atlasLibrary.langDb,
                                          renderingChoice=self.renderingChoice)

        self.setCentralWidget(self.fontGrid)
        self.fontItem = atlasLibrary.finder.substitute('sans-serif', 'Regular')

        self.updateFontGrid()

    closed = Signal()

    def closeEvent(self, event):
        r = super().closeEvent(event)
        if event.isAccepted():
            self.closed.emit()
        return r


SelectedCharacter = namedtuple('SelectedCharacter',
                               'codepoint char info encinfo generic fontitem')


def charToDict(charRes: SelectedCharacter, unicodeInfo: bool=False):
    """Convert a character to a dictionary for export.

    If unicodeInfo=True is passed, we also fetch unicode info about it."""

    result = {'codepoint': charRes.codepoint, 'char': charRes.char}
    if unicodeInfo:
        if charRes.info is not None:
            for field in UNICODE_DATA_FIELDS:
                if field == 'code':
                    continue
                result[field] = getattr(charRes.info, field)
        else:
            result.update(dict.fromkeys(UNICODE_DATA_FIELDS))

    if charRes.encinfo is not None:
        encinfo = charRes.encinfo

        if charRes.encinfo.encoding is not None:
            encoding = encinfo.encoding
            result['encoding'] = encoding.name
            result['multibyte'] = encoding.multibyte
            result['fixed'] = encoding.fixed
            result['charsize'] = encoding.charsize
            if encoding.morse:
                result['morse'] = encoding.alpha_codes.get(charRes.char)

        else:
            result['encoding'] = 'unknown'

        result['charvalue'] = encinfo.charvalue

        if encinfo.bytedata:
            result['bytes'] = binascii.b2a_base64(encinfo.bytedata).decode()
        else:
            result['bytes'] = None

    else:
        result['encoding'] = None

    if not charRes.generic:
        result['fontfamily'] = charRes.fontitem.family
        result['fontstyle'] = charRes.fontitem.style

    return result


def charToBytes(charRes: SelectedCharacter) -> bytes:
    """Convert charater to byte data, in the given encoding if one is passed."""
    if charRes.encinfo is None:
        return charRes.char.encode('utf8')
    elif (charRes.encinfo.encoding is not None and
            charRes.encinfo.encoding.morse):
        encoding = charRes.encinfo.encoding
        return encoding.alpha_codes.get(charRes.char).encode()
    else:
        return charRes.encinfo.bytedata


class CharSelectDialog(GlyphSelectorMixin, QtWidgets.QDialog):

    """A character selector dialog. Can be used withing TypeAtlas,
    within other Python programs, or from the command-line.

    You can configure the mode of the selector - whether to select
    multiple characters, which font to select from, do we use
    'select from font' mode for use with font-specific PUA symbols,
    what initial encoding to use, are encodings and symbol from
    font allowed to be changed.
    """

    def __init__(self, atlasLibrary: LibraryLike=None,
                       parent: QtWidgets.QWidget=None,
                       multipleChars: bool=False,
                       fontItem: fontlist.FontLike=None,
                       allowEncodings: bool=False, initialEncoding: str=None,
                       allowSymbolFromFontChange: bool=True,
                       selectFromFont: bool=False,
                       renderingChoice: FontRenderingChoice=None):

        super().__init__(parent)

        if atlasLibrary is None:
            atlasLibrary = GlyphAtlasLibrary.getInstance()

        self.resize(atlasLibrary.options.glyphAtlasSize)

        self.setWindowTitle(_("Select symbol or character"))
        self.setWindowIcon(getIcon('glyphatlas'))

        if renderingChoice is None:
            renderingChoice = FontRenderingChoice()

        self.atlasLibrary = atlasLibrary
        self.renderingChoice = renderingChoice

        self.fontGrid = fontgrid.FontGrid(atlasLibrary.charDb,
                                          fontDb=atlasLibrary.fontDb,
                                          langDb=atlasLibrary.langDb,
                                          renderingChoice=self.renderingChoice)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().addWidget(self.fontGrid)

        if fontItem is None:
            fontItem = atlasLibrary.finder.substitute('sans-serif', 'Regular')
            selectFromFont = False

        self.fontItem = fontItem
        self.selectFromFont = selectFromFont

        self.fontGrid.toggleCharBar(False)
        if not multipleChars:
            self.fontGrid.view.setSelectionMode(self.fontGrid.view.SingleSelection)

        if initialEncoding is not None:
            self.encoding = charsets.get_encoding(initialEncoding)

        self.updateFontGrid()

        self.multipleChars = multipleChars
        self.allowEncodings = allowEncodings
        self.initialEncoding = initialEncoding

        self.buttons = buttons = QtWidgets.QDialogButtonBox()
        buttons.addButton(_('&Select'), buttons.AcceptRole)
        buttons.addButton(_('&Cancel'), buttons.RejectRole)

        if allowEncodings:
            charsetBut = buttons.addButton(_("Change &encoding..."),
                                           buttons.ActionRole)
            self.charsetMenu = QtWidgets.QMenu()
            charsetBut.setMenu(self.charsetMenu)
            self.populateCharsetMenu()

        if allowSymbolFromFontChange:
            fontSymBut = buttons.addButton(_("Symbol from &font..."),
                                           buttons.ActionRole)
            fontSymBut.clicked.connect(self.selectSymbolFromFont)

            genericBut = buttons.addButton(_("Generic symbols"),
                                           buttons.ActionRole)
            genericBut.clicked.connect(self.selectGenericSymbol)

        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        self.layout().addWidget(buttons)

    def _indexToResult(self, index: QtCore.QModelIndex) -> SelectedCharacter:
        """Get which character is selected for the given index."""
        code = index.data(fontgrid.CharCodeRole)
        info = self.atlasLibrary.charDb.info_by_code.get(code)
        if self.charsetCodeInfo:
            encinfo = self.charsetCodeInfo.get(code)
        else:
            encinfo = None
        return SelectedCharacter(code, chr(code), info, encinfo,
                                 not self.selectFromFont,
                                 self.fontItem)

    @classmethod
    def selectChar(cls, *args, **kwargs) -> Optional[MaybeIterable[SelectedCharacter]]:
        """Open a dialog for selecting a character, and return one or more character
        depending on the mode."""

        self = cls(*args, **kwargs)
        if self.exec_():
            selection = self.fontGrid.view.selectionModel()
            current = selection.currentIndex()
            selected = selection.selectedIndexes()

            if not self.multipleChars:
                if current.isValid():
                    return self._indexToResult(current)
                elif selected:
                    return self._indexToResult(selected[0])
                else:
                    return None
            else:
                if selected:
                    return [self._indexToResult(index) for index in selected]
                elif current.isValid():
                    return [self._indexToResult(current)]
                else:
                    return None

        return None


class SplashScreen(QtWidgets.QSplashScreen):

    """GlyphAtlas splash screen."""

    msgAlignment = Qt.AlignLeft | Qt.AlignBottom

    def __init__(self):
        splashIcon = getImage('splash-glyphatlas')
        splashPixmap = splashIcon.pixmap(640, 480)
        super(SplashScreen, self).__init__(splashPixmap)


def glyphatlasStandalone():

    """Open GlyphAtlas. This is our main entry point."""

    install_locales()
    #load_samples()
    #load_nonfree()
    #load_glass()
    app = QtWidgets.QApplication(sys.argv)

    # Loading the style after the application ensures the colour
    # scheme is properly loaded.
    if configuredStyle:
        QtWidgets.QApplication.setStyle(configuredStyle)

    # Python 3 is a broken downgrade from 2, and no longer supports Unicode
    # paths until we set the filesystem encoding, and so opening the splash
    # crashes TypeAtlas. Instantiating a widget sets the correct encoding for
    # us. Once we have monkeypatched the FS encoding, starting the splash
    # screen is safe, and won't be racist against non-English characters
    # found in the pathname of the splash any more.
    #
    # I've always thought that the design decision to destroy Unicode support
    # in Python 3 was idiotic, but whatever, deal with it.
    # Or use a real language like PHP.
    QtWidgets.QWidget()

    splashW = SplashScreen()
    splashW.show()
    app.processEvents()

    library = GlyphAtlasLibrary.getInstance(splashW, fallback=False)
    splashW.finish(library.newGlyphAtlasWindow())

    print(_('Started %s, licensed under %s')
                % (_(proginfo.PROGRAM_NAME) + ' ' + proginfo.VERSION,
                   proginfo.LICENSE), file=sys.stderr)

    print(_('Using Qt %s toolkit with %s bindings')
                % (compat.QT_VERSION, compat.QT_BINDINGS), file=sys.stderr)

    app.exec_()


def _glyphatlasCliChooserLoopMain(app: QtWidgets.QApplication,
                                  args: 'argparse.Namespace',
                                  library: 'LibraryLike'):

    """The in-mainloop part of the CLI char selector.

    This is executed within the main loop"""
    import json

    fontItem = None

    if args.font:
        fontItem = library.finder.substitute(args.font, args.style)

    result = CharSelectDialog.selectChar(library,
                                         initialEncoding=args.charset,
                                         allowEncodings=args.charsets,
                                         multipleChars=args.multiple,
                                         allowSymbolFromFontChange=
                                            not args.no_symbol_mode_change,
                                         selectFromFont=args.symbol_from_font,
                                         fontItem=fontItem)

    if result is None:
        return

    encodedOutput = False

    if args.result_type == 'json':
        charToOutput = partial(charToDict, unicodeInfo=args.verbose)

    elif args.result_type == 'codepoint':
        charToOutput = lambda charRes: str(charRes.codepoint)

    elif args.result_type == 'raw':
        encodedOutput = True

        def charToOutput(charRes):
            result = charToBytes(charRes)
            if args.delimiter == '\0' and result == b'\0':
                result = b''
            elif args.delimiter == '\n' and result == b'\n':
                result = b''
            return result

    elif args.result_type in ['base64', 'hex', 'hqx', 'qp', 'uu']:

        b2a = getattr(binascii, 'b2a_' + args.result_type)

        def charToOutput(charRes):
            result = charToBytes(charRes)
            result = b2a(result).decode('ascii').strip()
            if args.result_type in ['base64', 'uu']:
                result = result.rstrip('\n')
            return result

    if encodedOutput:
        if args.output is not None:
            output = open(args.output, 'ab' if args.append else 'wb')
        else:
            output = sys.stdout.buffer

    else:
        if args.output is not None:
            output = open(args.output, 'at' if args.append else 'wt',
                          encoding='utf8')
        else:
            output = sys.stdout

    if args.multiple:
        resultdata = [charToOutput(c) for c in result]
    else:
        resultdata = charToOutput(result)

    if args.result_type == 'json':
        json.dump(resultdata, output, indent=args.indent)
        output.write('\n')

    else:
        delimiter = args.delimiter
        if encodedOutput:
            delimiter = delimiter.encode('ascii')

        if args.multiple:
            output.write(delimiter.join(resultdata))
        else:
            output.write(resultdata)

        output.write(delimiter)

    #app.exit(0)


def glyphatlasCliChooser():

    """Run the CLI character selector. This is another entry point."""

    global argparse

    install_locales()
    import argparse
    import codecs

    parser = argparse.ArgumentParser(description=_("Select symbol or character"))

    parser.add_argument('--output', '-o',
                        default=None, help=_("write the output to file"))
    parser.add_argument('--append', action='store_true',
                        help=_("append to output"))

    parser.add_argument('--result-type', '-r',
                        choices=['json', 'codepoint', 'raw',
                                 'base64', 'hex', 'hqx', 'qp', 'uu'],
                        default='json',
                        help=_("choose a result mode to return the character(s)"))

    parser.add_argument('--delimiter', '-d', '--separator',
                        default='\n',
                        type=lambda s: codecs.unicode_escape_decode(s)[0],
                        help=_("use the given character or escape code as "
                               "delimiter instead of newline"))
    parser.add_argument('--null', '-0',
                        dest='delimiter', action='store_const',
                        const='\0',
                        help=_("use null byte as constant"))

    parser.add_argument('--indent', default=4, type=int,
                        help=_("indent the json to the given level"))
    parser.add_argument('--no-indent', '-I',
                        action='store_const', const=0,
                        dest='indent',
                        help=_("do not indent the json"))

    parser.add_argument('--verbose', '-v', default=0,
                        help=_("include more information"), action='count')

    parser.add_argument('--multiple', '-m', action='store_true', default=False,
                        help=_("select multiple characters"))
    parser.add_argument('--no-multiple', '-M', '--no-m', '--single',
                        action='store_false', dest='multiple', default=False,
                        help=_("select a single character"))

    parser.add_argument('--charset', '--encoding', '-C',
                        help=_("return a character in a given character set"))
    parser.add_argument('--charsets', '-U', '--legacy-encodings',
                        action='store_true', default=False,
                        help=_("select legacy encodings and character sets"))
    parser.add_argument('--no-charsets', '-u', '--unicode',
                        action='store_false', dest='charsets', default=False,
                        help=_("get unicode result"))

    parser.add_argument('--no-symbol-mode-change',
                        '--generic-symbol-only', '--disallow-font-symbols',
                        action='store_true',
                        help=_("select only generic or font-specific characters"))

    parser.add_argument('--font', '-F', help=_("choose a font"))
    parser.add_argument('--style', '-S',  help=_("choose a style"),
                        default="Regular")
    parser.add_argument('--symbol-from-font', action='store_true',
                        help=_("default to symbols from the specified font"))

    parser.add_argument('--force-symbol-from-font', action='store_true',
                        help=_("only select symbols from the specified font"))

    parser.add_argument('--debug', action='store_true',
                        help=_("leave debug enabled"))

    args = parser.parse_args()

    if args.force_symbol_from_font:
        args.symbol_from_font = True
        args.no_symbol_mode_change = True

    if not args.debug:
        disable_debug()

    app = QtWidgets.QApplication([sys.argv[0]])

    # Loading the style after the application ensures the colour
    # scheme is properly loaded.
    if configuredStyle:
        QtWidgets.QApplication.setStyle(configuredStyle)

    # Python 3 is a broken downgrade from 2, and no longer supports Unicode
    # paths until we set the filesystem encoding, and so opening the splash
    # crashes TypeAtlas. Instantiating a widget sets the correct encoding for
    # us. Once we have monkeypatched the FS encoding, starting the splash
    # screen is safe, and won't be racist against non-English characters
    # found in the pathname of the splash any more.
    #
    # I've always thought that the design decision to destroy Unicode support
    # in Python 3 was idiotic, but whatever, deal with it.
    # Or use a real language like PHP.
    QtWidgets.QWidget()

    library = GlyphAtlasLibrary()
    #QtCore.QTimer.singleShot(0, partial(_glyphatlasCliChooserLoopMain,
    #                                     app, args, library))
    #app.exec_()
    _glyphatlasCliChooserLoopMain(app, args, library)


if __name__ == '__main__':
    glyphatlasStandalone()



# coding: utf-8

#  TypeAtlas Collection of International Sample Texts: Pangrams and others,
#  picked from Eat Glass collection of the Kermit project.
#
#     Phrases selected 2018 by Milko Krachounov, from the Eat Glass translations
#     from Kermit Project (a single phrase), contributed by various individuals.
#
#    They are distributed without any warranty.
#
#     There are no copyright notices with the UTF-8 sampler, in the original,
#     copies, or Pango re-use of the phrases, probably under the assumption
#     that the phrases fall below the threshold of originality for copyright.
#
#
###########################################################################
#
#   2) Sources, their copyright and licenses:
#      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
#  * Eat Glass translations from Kermit Project. Written and copyrighted by the
#    contributors, Frank da Cruz, the Kermit Project, Columbia University and
#    Ethan Mollick
#
#    Copy of the UTF-8 sampler: <http://www.columbia.edu/~fdc/utf8.html>
#
#    No notice on copyright and copying was found, but the collection
#    was not copied here. Only several (eight, including the English) individual
#    phrases, contributed by third parties, were included. That includes
#    six phrases also included by Pango. Two phrases are in this file.

# Any contributions to this file by Milko Krachounov are quintuple-licensed under GPL 3,
#      LGPL 2, MIT/X11, CC-BY-SA 3.0 and CC-BY-SA 4.0 (just to be on the safe
#      side). (If this is too restrictive, I can be persuaded to grant more).
#      See the notice in samples_language.py on the matter.


from typeatlas.samples import add_sample, use_fallback, use_display_scripts
from typeatlas.samples import PHRASE, PANGRAM, PANGRAM_PARTIAL_ACCENTS
from typeatlas.samples import PANGRAM_LIMITED_ACCENTS, PARTIAL_PANGRAM, PERFECT_PANGRAM
from typeatlas.samples import KOREAN_PANGRAM, KOREAN_CONSONANT_PANGRAM
from typeatlas.samples import YONG_CHARACTER, SYLLABLE_PANGRAM
from typeatlas.samples import PRIO_PREFER, PRIO_WITHHOLD


EAT_GLASS_ENGLISH = "I can eat glass and it doesn't hurt me."
EAT_GLASS_SOURCE = 'UTF-8 Sampler - The Kermit Project <http://kermitproject.org/utf8.html>'

# This should be second
add_sample('ha', "إِنا إِىَ تَونَر غِلَاشِ كُمَ إِن غَمَا لَافِىَا.",
           script='Arab',
           sources=[EAT_GLASS_SOURCE],
           origin='Malami Buba, Tom Gewecke.',
           english=EAT_GLASS_ENGLISH,
           langpos=1, langscriptpos=0)



# -*- coding: utf-8 -*-
#
#    TypeAtlas Character Information
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#


"""Access unicode data, allowing the download of all the unicode data from
unicode.org, and falling back to Python's unicodedata when that's not
available.

Nothing should be downloaded without user's consent here.
"""

from typeatlas.util import OrderedSet, N_, warnmsgf, generic_type
from typeatlas import proginfo, rangemath
from collections import namedtuple, OrderedDict, defaultdict
from collections.abc import Callable
from itertools import chain
import typeatlas
import unicodedata
import urllib.parse
import urllib.request
import shutil
import bisect
import sys
import os
import os.path
import io
import re
import copy

SequenceOf = generic_type('Sequence')
TupleOf = generic_type('Tuple')
IterableOf = generic_type('Iterable')
IteratorOf = generic_type('Iterator')
Optional = generic_type('Optional')
Union = generic_type('Union')


UNICODE_DOWNLOAD_BASE = 'https://www.unicode.org/Public/UCD/latest/ucd/'
UNICODE_CACHE_PATH = os.path.join(proginfo.CACHE_DIR, 'unicode')
UNICODE_SEARCH_PATH = [UNICODE_CACHE_PATH, '/usr/share/unicode']
UNICODE_NAMESLIST_FILE = 'NamesList.txt'
UNICODE_NAMEALIASES_FILE = 'NameAliases.txt'
UNICODE_DATA_FILE = 'UnicodeData.txt'
UNICODE_BLOCKS_FILE = 'Blocks.txt'
UNICODE_SCRIPTS_FILE = 'Scripts.txt'
UNICODE_VALUE_ALIASES_FILE = 'PropertyValueAliases.txt'
UNICODE_HANGUL_SYLLABLE_FILE = 'HangulSyllableType.txt'

UCSUR_DOWNLOAD_BASE = r'https://www.kreativekorp.com/ucsur/UNIDATA/'
UCSUR_CACHE_PATH = os.path.join(proginfo.CACHE_DIR, 'ucsur')
UCSUR_SEARCH_PATH = [os.path.join(proginfo.CACHE_DIR, 'ucsur')]

UNOFFICIAL_REGISTRIES = {
    'UCSUR': dict(download_base=UCSUR_DOWNLOAD_BASE,
                  cache_path=UCSUR_CACHE_PATH,
                  search_path=UCSUR_SEARCH_PATH,
                  name=N_('ConScript Unicode Registry (unofficial)'))

}



PLANE_SIZE = 0x10000
PLANE_END = PLANE_SIZE - 1

PLANE_COUNT = 17

LOW_SURROGATE_START = LOW_SURROGATE_FIRST = 0xD800
LOW_SURROGATE_STOP = 0xDC00
LOW_SURROGATE_LAST = LOW_SURROGATE_STOP - 1

HIGH_SURROGATE_START = HIGH_SURROGATE_FIRST = LOW_SURROGATE_STOP
HIGH_SURROGATE_STOP = 0xE000
HIGH_SURROGATE_LAST = HIGH_SURROGATE_STOP - 1


SURROGATE_START = SURROGATE_FIRST = LOW_SURROGATE_START
SURROGATE_STOP = HIGH_SURROGATE_STOP
SURROGATE_LAST = SURROGATE_STOP - 1

# Note that sys.maxunicode / 0x10FFFF is a non-character, but is valid
# in interchange, so we include it in the result.
UNICODE_START = UNICODE_FIRST = 0
UNICODE_LAST = sys.maxunicode
#UNICODE_LAST = 0x10FFFF
UNICODE_STOP = sys.maxunicode + 1

NON_CHARACTER_CODES = (set(plane * PLANE_SIZE + PLANE_END - 1
                           for plane in range(PLANE_COUNT)) |
                       set(plane * PLANE_SIZE + PLANE_END
                           for plane in range(PLANE_COUNT)) |
                       set(range(0xFDD0, 0xFDEF + 1)))

NON_CHARACTERS = set(map(chr, NON_CHARACTER_CODES))
ZEROWIDTH = frozenset('\x00\u034F\u200B\u200F\u2028\u2029'
                      '\u202A\u202E\u2060\u2063')


_UNSPECIFIED = object()


_empty_file = lambda: io.StringIO('')


SYMBOL_FOR_PREFIX = 'SYMBOL FOR '

UnicodePlane = namedtuple('UnicodePlane', 'start end number description')

PLANES = {
    0: N_('Basic Multilingual Plane'),
    1: N_('Supplementary Multilingual Plane'),
    2: N_('Supplementary Ideographic Plane'),
    14: N_('Supplementary Special-purpose Plane'),
    15: N_('Supplementary Private Use Area Plane A'),
    16: N_('Supplementary Private Use Area Plane B'),
}

CATEGORIES = {
    'Lu': (N_('Letter'), N_('Uppercase')),
    'Ll': (N_('Letter'), N_('Lowercase')),
    'Lt': (N_('Letter'), N_('Titlecase')),
    'Mn': (N_('Mark'), N_('Non-Spacing')),
    'Mc': (N_('Mark'), N_('Spacing Combining')),
    'Me': (N_('Mark'), N_('Enclosing')),
    'Nd': (N_('Number'), N_('Decimal Digit')),
    'Nl': (N_('Number'), N_('Letter')),
    'No': (N_('Number'), N_('Other')),
    'Zs': (N_('Separator'), N_('Space')),
    'Zl': (N_('Separator'), N_('Line')),
    'Zp': (N_('Separator'), N_('Paragraph')),
    'Cc': (N_('Other'), N_('Control')),
    'Cf': (N_('Other'), N_('Format')),
    'Cs': (N_('Other'), N_('Surrogate')),
    'Co': (N_('Other'), N_('Private Use')),
    'Cn': (N_('Other'), N_('Not Assigned')),

    'Lm': (N_('Letter'), N_('Modifier')),
    'Lo': (N_('Letter'), N_('Other')),
    'Pc': (N_('Punctuation'), N_('Connector')),
    'Pd': (N_('Punctuation'), N_('Dash')),
    'Ps': (N_('Punctuation'), N_('Open')),
    'Pe': (N_('Punctuation'), N_('Close')),
    'Pi': (N_('Punctuation'), N_('Initial quote')),
    'Pf': (N_('Punctuation'), N_('Final quote')),
    'Po': (N_('Punctuation'), N_('Other')),
    'Sm': (N_('Symbol'), N_('Math')),
    'Sc': (N_('Symbol'), N_('Currency')),
    'Sk': (N_('Symbol'), N_('Modifier')),
    'So': (N_('Symbol'), N_('Other')),
}

CATEGORY_ICONS = {
    'Lu': 'character-letter',
    'Ll': 'character-letter',
    'Lt': 'character-letter',
    'Mn': 'character-mark',
    'Mc': 'character-mark',
    'Me': 'character-mark',
    'Nd': 'character-number',
    'Nl': 'character-number',
    'No': 'character-number',
    'Zs': 'character-separator',
    'Zl': 'character-separator',
    'Zp': 'character-separator',
    'Cc': 'character-control',
    'Cf': 'character-format',
    'Cs': 'character-surrogate',
    'Co': 'character-private-use',
    'Cn': 'character-unassigned',
    'Lm': 'character-letter',
    'Lo': 'character-letter',
    'Pc': 'character-punctuation',
    'Pd': 'character-punctuation',
    'Ps': 'character-punctuation',
    'Pe': 'character-punctuation',
    'Pi': 'character-punctuation',
    'Pf': 'character-punctuation',
    'Po': 'character-punctuation',
    'Sm': 'character-symbol',
    'Sc': 'character-symbol',
    'Sk': 'character-symbol',
    'So': 'character-symbol',

    'L': 'character-letter',
    'M': 'character-mark',
    'N': 'character-number',
    'Z': 'character-separator',
    'P': 'character-punctuation',
    'S': 'character-symbol',
    'C': 'character-private-use',
}

BIDI_CATEGORIES = {
    'L': N_('Left-to-Right'),
    'LRE': N_('Left-to-Right Embedding'),
    'LRO': N_('Left-to-Right Override'),
    'R': N_('Right-to-Left'),
    'AL': N_('Right-to-Left Arabic'),
    'RLE': N_('Right-to-Left Embedding'),
    'RLO': N_('Right-to-Left Override'),
    'PDF': N_('Pop Directional Format'),
    'EN': N_('European Number'),
    'ES': N_('European Number Separator'),
    'ET': N_('European Number Terminator'),
    'AN': N_('Arabic Number'),
    'CS': N_('Common Number Separator'),
    'NSM': N_('Non-Spacing Mark'),
    'BN': N_('Boundary Neutral'),
    'B': N_('Paragraph Separator'),
    'S': N_('Segment Separator'),
    'WS': N_('Whitespace'),
    'ON': N_('Other Neutrals'),
}

UNICODE_DATA_FIELDS = tuple([
    'code', 'name', 'category', 'combining',
    'bidirectional', 'decomposition', 'decimal',
    'digit', 'numeric', 'mirrored', 'old_name', 'comment',
    'uppercase', 'lowercase', 'titlecase',
])


CHAR_INFO_FIELD_DEFAULTS = {
    'category': 'Cn',
    'aliases': (),
    'formalaliases': (),
    'variations': (),
    'crossrefs': (),
}

formal_alias = '\u203B'
cross_reference = '\u2192'

# Unicode blocks
uniblock_line_rgx = re.compile(r'(?i)^\s*([0-9a-f]+)\.\.([0-9a-f]+);\s*(\S.*)')

# Names list
nameblock_line_rgx = re.compile(r'(?i)^@@\s+([0-9a-f]+)\s+(.*\S)\s+([0-9a-f]+)')
subheader_line_rgx = re.compile(r'(?i)^@\s+(\S.*)')
code_name_line_rgx = re.compile(r'^(?i)([0-9a-f]+)\s+(\S.*)')
extra_rgx = re.compile(r'\s([=\+x%~])\s+(\S.*)')

range_first_rgx = re.compile(r'^<(.*), First>')
range_last_rgx = re.compile(r'^<(.*), Last>')


NAMELIST_PREFIXES = {
    '=': 'aliases',
    '%': 'formalaliases',
    'x': 'crossrefs',
    '~': 'variations',
}

MEANING_TO_OLD_NAME = {
    'LINE FEED': 'LINE FEED (LF)',
    'HORIZONTAL TABULATION': 'CHARACTER TABULATION',
    'VERTICAL TABULATION': 'LINE TABULATION',
    'FORM FEED': 'FORM FEED (FF)',
    'CARRIAGE RETURN': 'CARRIAGE RETURN (CR)',
    'FILE SEPARATOR': 'INFORMATION SEPARATOR FOUR',
    'GROUP SEPARATOR': 'INFORMATION SEPARATOR THREE',
    'RECORD SEPARATOR': 'INFORMATION SEPARATOR TWO',
    'UNIT SEPARATOR': 'INFORMATION SEPARATOR ONE',
}

OLD_NAME_MEANING = {v: k for k, v in MEANING_TO_OLD_NAME.items()}


CHAR_INFO_FIELDS = OrderedSet(UNICODE_DATA_FIELDS)
CHAR_INFO_FIELDS.update(['display', 'block', 'subheader'])
CHAR_INFO_FIELDS.update(NAMELIST_PREFIXES.values())
CHAR_INFO_FIELDS.add('registry')


UnicodeBlock = namedtuple('UnicodeBlock', 'start end name')
CharRange = namedtuple('CharRange', 'start end info')

BRAILLE_BLOCKS = (UnicodeBlock(0x2800, 0x28FF, 'Braille Patterns'), )
BRAILLE_RANGE = rangemath.OrdinalRange(0x2800, 0x28FF)


class CharInfo(object):

    """The info about a given characters.

    This is initialised using the CHAR_INFO_FIELDS fields, which
    includes the standard unicode data fields (UNICODE_DATA_FIELDS),
    and some additional ones."""

    __slots__ = tuple(CHAR_INFO_FIELDS)

    def __init__(self, *args, **kwargs):
        self.update(*args, _initial=True, **kwargs)

    def __repr__(self):
        code = self.code
        if code is None:
            code = -1
        return '<CharInfo 0x%X %r; at 0x%x>' % (code, self.name,
                                                id(self))

    def merge(self, other: 'CharInfo'):
        """Merge the info from other char info into this one.
        This changes all fields of this character that are
        the default, to the one from the other font info.
        """
        for key in CHAR_INFO_FIELDS:
            value = getattr(other, key)
            if value != CHAR_INFO_FIELD_DEFAULTS.get(key):
                setattr(self, key, value)

    def update(self, *args, _initial=False, **kwargs):
        """Update this character info, as if the constructor
        was called again."""
        if len(args) > len(CHAR_INFO_FIELDS):
            raise TypeError("update takes %r positional argument "
                                      "but %r were given"
                                    % (len(CHAR_INFO_FIELDS) + 1,
                                       len(args) + 1))

        for value, key in zip(args, CHAR_INFO_FIELDS):
            if key in kwargs:
                raise TypeError("update got multiple values for argument %r"
                                    % (key, ))
            kwargs[key] = value

        for key, value in kwargs.items():
            if key not in CHAR_INFO_FIELDS:
                raise TypeError("update got an unexpected keyword argument %r"
                                    % (key, ))

            setattr(self, key, value)

        if _initial:
            for key in CHAR_INFO_FIELDS:
                if not hasattr(self, key):
                    setattr(self, key, CHAR_INFO_FIELD_DEFAULTS.get(key))

    def category_name(self, template: str='{category}, {subcategory}',
                            translate: Callable=N_) -> str:
        """Return the name of the category of the font. By default,
        it is formatted as '{category}, {subcategory}', but you can
        override that.

        You can provide a translation function with translate=_. One
        should pass translate, as the returned value is not guaranteed
        to be translatable as-is.
        """
        category = self.category
        if category not in CATEGORIES:
            category = 'Cn'

        cat, subcat = CATEGORIES.get(category, (N_('Unknown'), N_('Unknown')))
        return template.format(category=translate(cat),
                               subcategory=translate(subcat))

    @property
    def category_icon(self) -> str:
        """Return the category icon name for the function. It should
        represent a file included with typeatlas, without the extension,
        describing the given category of character"""
        try:
            return CATEGORY_ICONS[self.category]
        except KeyError:
            pass

        try:
            return CATEGORY_ICONS[self.category[0]]
        except (KeyError, IndexError, TypeError):
            pass

        try:
            return CATEGORY_ICONS['Co']
        except KeyError:
            pass

        return None


class CharacterDatabase(object):

    """The database of Unicode characters. This holds the unicode data.

    Objects of this class can also implement alternative registries.
    """

    instance = None

    def __init__(self, download_base: str=UNICODE_DOWNLOAD_BASE,
                       cache_path: SequenceOf[str]=UNICODE_CACHE_PATH,
                       search_path: SequenceOf[str]=UNICODE_SEARCH_PATH,
                       name: str=N_('Unicode Character Database'),
                       enabled: bool=True,
                       secondary_registries: 'SequenceOf[CharacterDatabase]'=()):

        self.populated = False

        self.info_by_code = {}
        self.info_by_name = {}
        self.info_by_old_name = {}
        self.infos_by_name = defaultdict(list)

        self.info_ranges = []
        self.info_range_starts = []

        self.unicode_blocks = []
        self.unicode_block_starts = []
        self.scripts = {}
        self.scriptorder = []
        self.script_blocks = []
        self.script_block_starts = []
        self.value_aliases = defaultdict(dict)
        self.value_aliases_all = defaultdict(lambda: defaultdict(list))
        self.alias_values = defaultdict(dict)
        self.download_base = download_base
        self.cache_path = cache_path
        self.search_path = search_path

        self.combining_chars = set()

        self.name = name
        self.enabled = enabled
        self.secondary_registries = list(secondary_registries)

    @classmethod
    def get_instance(cls, populated: bool=False) -> 'CharacterDatabase':
        """Return the singleton instance of the character database, or create
        one if there is not one created yet.

        If populated=True is passed, the instance is required to be
        populated."""
        if cls.instance is None:
            cls.instance = cls()
        if populated:
            cls.instance.ensure_populated()
        return cls.instance

    getInstance = get_instance

    def add_registry_custom(self, registry: 'CharacterDatabase'):
        """Add a custom additional unofficial registry of Unicode
        characters."""
        self.secondary_registries.append(registry)

    def add_registry(self, name: str):
        """Add a custom additional unofficial registry of Unicode
        characters by name, using the known registries from
        UNOFFICIAL_REGISTRIES."""
        cls = type(self)
        registry = cls(**UNOFFICIAL_REGISTRIES[name])
        self.add_registry_custom(registry)

    def add_registries(self, names: IterableOf[str]=None):
        """Add all known or all specified unofficial unicode registries."""
        if names is None:
            names = list(UNOFFICIAL_REGISTRIES)
        for name in names:
            self.add_registry(name)

    def registries(self) -> 'IteratorOf[CharacterDatabase]':
        yield self
        for registry in self.secondary_registries:
            yield from registry.registries()

    def downloadables(self, deep: bool=False) -> IteratorOf[TupleOf[str, str]]:
        """Return pairs of URL, destination path for all the files that
        need to be downloaded for this class to function with latest
        unicode data.

        If deep=True is passed, any added unofficial registries are also
        iterated over.
        """

        yield self.downloadable(UNICODE_BLOCKS_FILE)
        yield self.downloadable(UNICODE_SCRIPTS_FILE)
        yield self.downloadable(UNICODE_DATA_FILE)
        yield self.downloadable(UNICODE_NAMESLIST_FILE)
        #yield self.downloadable(UNICODE_NAMEALIASES_FILE)
        yield self.downloadable(UNICODE_VALUE_ALIASES_FILE)
        #yield self.downloadable(UNICODE_HANGUL_SYLLABLE_FILE)
        if not deep:
            return
        for registry in self.secondary_registries:
            yield from registry.downloadables(deep)

    def downloadable(self, filename: str) -> TupleOf[str, str]:
        """Given the provided filename, return a tuple of the URL
        where to download, and the path where to download it to."""
        if not os.path.exists(self.cache_path):
            os.makedirs(self.cache_path)
        url = urllib.parse.urljoin(self.download_base, filename)
        destpath = os.path.join(self.cache_path, filename)
        return url, destpath

    def download_file(self, filename: str):
        """Download the given filename from the correct URL to the
        correct destination path"""
        url, destpath = self.downloadable(filename)
        with urllib.request.urlopen(url) as src:
            with open(destpath, 'wb') as dst:
                shutil.copyfileobj(src, dst)

    def search(self, filename: str, download: bool=False,
                                    default=_UNSPECIFIED) -> Optional[str]:
        """Find the given unicode filename in all potential location
        where the operating system can store it. This will only work
        on some operating systems."""
        for path in self.search_path:
            filepath = os.path.join(path, filename)
            if os.path.exists(filepath):
                return filepath

        if download:
            try:
                self.download_file(filename)
            except OSError:
                if default is _UNSPECIFIED:
                    raise
            filepath = os.path.join(self.cache_path, filename)
            if os.path.exists(filepath):
               return filepath

        if default is not _UNSPECIFIED:
            return default
        raise KeyError(filename)

    def get(self, code: int, autofill: bool=False,
                             fallback: bool=False) -> CharInfo:
        """Get character info by code.

        If autofill is True, basic information may be provided even
        if the character is unknown.

        If fallback is True, data from Python's unicodedata may be
        returned if we don't have unicode data.
        """
        result = None
        if self.info_by_code:
            result = self.info_by_code.get(code)
        else:
            try:
                char = chr(code)
                result = CharInfo(code=code,
                                  name=unicodedata.name(char),
                                  category=unicodedata.category(char),
                                  combining=unicodedata.combining(char),
                                  bidirectional=unicodedata.bidirectional(char),
                                  decomposition=unicodedata.decomposition(char),
                                  decimal=unicodedata.decimal(char),
                                  digit=unicodedata.digit(char),
                                  numeric=unicodedata.numeric(char),
                                  mirrored=unicodedata.mirrored(char),
                                  registry=self)
            except ValueError:
                pass

        if result is None:
            i = bisect.bisect(self.info_range_starts, code) - 1
            if 0 <= i < len(self.info_ranges):
                charrange = self.info_ranges[i]
                if charrange.start <= code <= charrange.end:
                    result = copy.copy(charrange.info)
                    result.code = code
                    try:
                        result.name = unicodedata.name(chr(code))
                    except ValueError:
                        pass

        if result is None and fallback:
            for registry in self.secondary_registries:
                result = registry.get(code, autofill=False)
                if result is not None:
                    break

        if result is None and autofill:
            result = CharInfo(code=code,
                              block=self.find_block(code, None))
        return result

    def get_secondary(self, code: int) -> SequenceOf[CharInfo]:
        """Get secondary information from other registries"""
        results = []
        for registry in self.secondary_registries:
            extra = registry.get(code, autofill=False)
            if extra is not None:
                results.append(extra)
        return results

    def getall(self, code, autofill: bool=False) -> SequenceOf[CharInfo]:
        """Get the information for the character from all registries.

        If autofill is True, basic information may be provided even
        if the character was not found in any of the registries.
        """
        results = []
        main = self.get(code, autofill=False)
        if main is not None:
            results.append(main)

        results.extend(self.get_secondary(code))

        if not results and autofill:
            results.append(CharInfo(code=code,
                                    block=self.find_block(code, None)))
        return results

    def lookup(self, name: str) -> CharInfo:
        """Get character info by name."""
        if self.info_by_code:
            return self.info_by_name.get(name)
        try:
            code = ord(unicodedata.lookup(name))
        except KeyError:
            return None
        else:
            return self.get(code, autofill=True)

    def lookup_many(self, name: str) -> SequenceOf[CharInfo]:
        """Get all characters with a given name."""
        if self.info_by_code:
            return tuple(self.infos_by_name.get(name, ()))
        else:
            return (self.lookup(name), )

    def get_plane(self, code: int,
                        default_description: str
                            =N_('Unassigned plane')) -> UnicodePlane:
        """Get the unicode plane for the character"""
        plane = code // PLANE_SIZE
        description = PLANES.get(plane, default_description)
        return UnicodePlane(plane * PLANE_SIZE, plane * PLANE_SIZE + PLANE_END,
                            plane, description)

    def find_block(self, code: int,
                         default=_UNSPECIFIED) -> Optional[UnicodeBlock]:
        """Find the block of the character."""
        i = bisect.bisect(self.unicode_block_starts, code) - 1
        if 0 <= i < len(self.unicode_blocks):
            block = self.unicode_blocks[i]
            if block.start <= code <= block.end:
                return block
        if default is _UNSPECIFIED:
            raise LookupError(code)
        return default

    def find_script_block(self, code: int, default=_UNSPECIFIED
                          ) -> Optional[UnicodeBlock]:
        """Find the script block of the letter."""
        i = bisect.bisect(self.script_block_starts, code) - 1
        if 0 <= i < len(self.script_blocks):
            block = self.script_blocks[i]
            if block.start <= code <= block.end:
                return block
        if default is _UNSPECIFIED:
            raise LookupError(code)
        return default

    def find_script(self, code: int, default='Unknown') -> Optional[str]:
        """Find the script of the letter."""
        i = bisect.bisect(self.script_block_starts, code) - 1
        if 0 <= i < len(self.script_blocks):
            block = self.script_blocks[i]
            if block.start <= code <= block.end:
                return block.name
        if default is _UNSPECIFIED:
            raise LookupError(code)
        return default

    def find_script_code(self, code: int, default='Zzzz') -> str:
        """Find the ISO 15924 code for the script of the given letter."""
        script = self.find_script(code, None)
        if script is None:
            if default is _UNSPECIFIED:
                raise LookupError(code)
            return default
        return self.get_value_alias('sc', script)

    def find_script_name(self, langdb: 'typeatlas.langutil.LanguageDatabase',
                               code: int, which=None,
                               default=None) -> Optional[str]:
        """Find the name of the script of the given letter.

        Can be the localized, non-localized, or native name. See
        the documentation for langutil.LanguageDatabase.
        """
        script = self.find_script(code, None)

        if script is None:
            if default is not None:
                return default
            if default is _UNSPECIFIED:
                raise LookupError(code)
            script = 'Unknown'

        return langdb.script_name(script, which, chardb=self)

    def find_scripts(self, codes: Union[IterableOf[int], str],
                           default='Unknown') -> IteratorOf[str]:
        """Find the scripts of the letters. A script for every
        letter is returned."""

        block = None

        if isinstance(codes, str):
            codes = map(ord, codes)

        for code in codes:
            if block is not None and block.start <= code <= block.end:
                yield block.name
                continue

            i = bisect.bisect(self.script_block_starts, code) - 1
            if 0 <= i < len(self.script_blocks):
                block = self.script_blocks[i]
                if block.start <= code <= block.end:
                    yield block.name
                    continue

                else:
                    block = None

            yield default

    def get_value_alias(self, category: str, value: str,
                              *other_values: str) -> str:
        """Get the alias for property value. For example, in
        category 'blk', 'CJK' will be alias for the value
        'CJK_Unified_Ideographs'"""
        if other_values:
            return self.get_values_aliases(category, (value, ) + other_values)
        return next(self.get_values_aliases(category, [value]))

    def get_values_aliases(self, category: str,
                                 values: IterableOf[str]) -> IteratorOf[str]:
        """Get the aliases for property values. See get_value_alias."""
        for value in values:
            yield self.value_aliases[category].get(value, value)

    def get_alias_value(self, category: str, value: str,
                              *other_values: str) -> str:
        """Get the property value for alias. See get_value_alias."""
        if other_values:
            return self.get_aliases_values(category, (value, ) + other_values)
        return next(self.get_aliases_values(category, [value]))

    def get_aliases_values(self, category: str,
                                 values: IterableOf[str]) -> IteratorOf[str]:
        """Get the values for aliases"""
        for value in values:
            yield self.alias_values[category].get(value, value)

    def combining(self, code: int) -> bool:
        """Return True if the character is combining."""
        chars = self.combining_chars
        if chars:
            return code in chars
        else:
            return unicodedata.combining(chr(code))

    def east_asian_width(self, code: int) -> str:
        """Get the east asian width. For a given width, you can get the
        meaning of each width using self.get_alias_value('ea', width)
        """
        # FIXME: We do not parse EastAsianWidth.txt
        return unicodedata.east_asian_width(chr(code))

    def guess_width(self, code: int, controlwidth: int=0, *,
                          mode='terminal') -> int:
        """Return a guess about the width of the given unicode
        codepoint. This is useful, for example, for terminal character
        width."""

        if code in ZEROWIDTH:
            return 0

        combining = self.combining_chars
        if code in ZEROWIDTH or code in combining:
            return 0

        char = chr(code)
        if not combining and unicodedata.combining(char):
            return 0

        if unicodedata.east_asian_width(char) in 'WF':
            return 2

        infos = self.info_by_code
        if code in infos:
            if infos[code].category == 'Cc':
                return controlwidth
        elif unicodedata.category(char) == 'Cc':
            return controlwidth

        return 1

    def guess_text_width(self, text: int, controlwidth: int=0, *,
                               mode='terminal') -> int:
        """Return a guess about the width of the given character(s)."""
        return sum(self.guess_code_width(ord(char), controlwidth)
                   for char in text)

    @classmethod
    def braille_blocks(cls) -> SequenceOf[UnicodeBlock]:
        """Get the Braille Patterns unicode blocks."""
        return BRAILLE_BLOCKS

    @classmethod
    def braille_range(cls) -> 'typeatlas.rangemath.RangeBase':
        """Get the Braille Patterns unicode range."""
        return BRAILLE_RANGE

    def open(self, filename: str, download: bool=False,
                   default=_UNSPECIFIED) -> Optional[io.TextIOBase]:
        """Find the given unicode filename and open it."""
        path = self.search(filename, download=download,
                           default=None
                                    if default is not _UNSPECIFIED
                                    else _UNSPECIFIED)
        if path is None:
            return default

        try:
            return open(path, 'rt', encoding='utf8')
        except OSError:
            if default is _UNSPECIFIED:
                raise
        else:
            return default

    def download(self, deep: bool=False):
        """Download all files rquired by this class. If deep=True is
        passed, unofficial registries attached to this are also
        downloaded."""
        self.download_file(UNICODE_BLOCKS_FILE)
        self.download_file(UNICODE_SCRIPTS_FILE)
        self.download_file(UNICODE_DATA_FILE)
        self.download_file(UNICODE_NAMESLIST_FILE)
        #self.download_file(UNICODE_NAMEALIASES_FILE)
        self.download_file(UNICODE_VALUE_ALIASES_FILE)
        #self.download_file(UNICODE_HANGUL_SYLLABLE_FILE)
        if deep:
            for registry in self.secondary_registries:
                registry.download(deep=deep)

    def ensure_populated(self, download: bool=False, deep: bool=False):
        """Ensure the registry is populated it, by reading any unicode
        files found if they weren't read already.

        If download=True is passed, the registries are download. If
        deep=True is also passed, the unofficial registries are also
        populated and/or downloaded."""
        if not self.populated:
            self.populate(download=download)
        if deep:
            for registry in self.secondary_registries:
                registry.ensure_populated(download=download, deep=deep)

    def populate(self, download: bool=False, deep: bool=False):
        """Populate the registry by reading any unicode files found.

        If download=True is passed, the registries are download. If
        deep=True is also passed, the unofficial registries are also
        populated and/or downloaded."""
        self.populate_blocks(download=download)
        self.populate_scripts(download=download)
        self.populate_unicode_data(download=download)
        self.populate_nameslist(download=download)
        #self.populate_namealiases(download=download)
        self.populate_value_aliases(download=download)
        #self.populate_hangul_syllables(download=download)

        self.populated = True

        if deep:
            for registry in self.secondary_registries:
                registry.populate(download=download, deep=deep)

    def populate_blocks(self, download: bool=False):
        """Populate the unicode blocks by reading Blocks.txt.

        It will be downloaded if download=True is passed.
        """

        with self.open(UNICODE_BLOCKS_FILE, default=_empty_file(),
                                            download=download) as src:
            blocks = []
            for line in src:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue
                match = uniblock_line_rgx.search(line)
                if match is not None:
                    start, end, name = match.groups()
                    blocks.append(UnicodeBlock(int(start, 16),
                                               int(end, 16), name))

        self.unicode_blocks = blocks
        self.unicode_block_starts = [block.start for block in blocks]

    def populate_scripts(self, download: bool=False):
        """Populate the unicode scripts by reading Scripts.txt.

        It will be downloaded if download=True is passed.
        """

        with self.open(UNICODE_SCRIPTS_FILE, default=_empty_file(),
                                             download=download) as src:
            script_blocks = []
            scripts = {}
            for line in src:
                line = line.partition('#')[0].strip()
                if not line:
                    continue
                coderange, sep, script = line.partition(';')
                if not sep:
                    continue

                coderange = coderange.strip()
                script = script.strip()

                start, sep, end = coderange.partition('..')
                if not sep:
                    end = start
                block = UnicodeBlock(int(start, 16), int(end, 16), script)
                script_blocks.append(block)
                scripts.setdefault(block.name, []).append(block)
            script_blocks.sort()

        self.script_blocks = script_blocks
        self.script_block_starts = [block.start for block in script_blocks]
        self.scripts = scripts
        self.scriptorder = sorted(scripts)

    def populate_unicode_data(self, download: bool=False):
        """Populate the unicode data by reading UnicodeData.txt.

        It will be downloaded if download=True is passed.
        """

        self.infos_by_name.clear()

        info_range_starts = OrderedDict()
        info_range_ends = OrderedDict()

        with self.open(UNICODE_DATA_FILE, default=_empty_file(),
                                          download=download) as src:
            for line in src:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue

                values = line.split(';')

                if len(values) < 2:
                    continue

                info = CharInfo(registry=self,
                                **dict(zip(UNICODE_DATA_FIELDS, values)))
                info.code = int(info.code, 16)

                if info.uppercase:
                    info.uppercase = int(info.uppercase, 16)
                else:
                    info.uppercase = None

                if info.lowercase:
                    info.lowercase = int(info.lowercase, 16)
                else:
                    info.lowercase = None

                if info.titlecase:
                    info.titlecase = int(info.titlecase, 16)
                else:
                    info.titlecase = None

                if (info.combining and
                    info.combining.isdigit() and
                    info.combining.isascii()):
                        info.combining = \
                            int(info.combining)

                info.category = sys.intern(info.category)
                info.bidirectional = sys.intern(info.bidirectional)
                existing = self.info_by_code.get(info.code)
                if existing is not None:
                    existing.merge(info)
                    info = existing

                if info.name.startswith('<') and info.name != '<control>':
                    match = range_first_rgx.search(info.name)
                    if match is not None:
                        info_range_starts[match.group(1)] = info
                        continue

                    match = range_last_rgx.search(info.name)
                    if match is not None:
                        info_range_ends[match.group(1)] = info
                        continue

                self.info_by_code[info.code] = info
                if info.name:
                    if info.name != '<control>':
                        self.info_by_name[info.name] = info
                    self.infos_by_name[info.name].append(info)
                if info.old_name:
                    self.info_by_old_name[info.old_name] = info
                    self.infos_by_name[info.old_name].append(info)

        self.combining_chars.clear()

        for info in self.info_by_code.values():
            if info.name and info.name.startswith(SYMBOL_FOR_PREFIX):
                symbol_for = info.name[len(SYMBOL_FOR_PREFIX):]
                if symbol_for in self.info_by_old_name:
                    self.info_by_old_name[symbol_for].display = info.code
                else:
                    symbol_for = MEANING_TO_OLD_NAME.get(symbol_for)
                    if symbol_for in self.info_by_old_name:
                        self.info_by_old_name[symbol_for].display = info.code

            if info.combining:
                self.combining_chars.add(info.code)

        for old_name, info in self.info_by_old_name.items():
            if old_name not in self.info_by_name:
                self.info_by_name[old_name] = info

        del self.info_ranges[:]
        del self.info_range_starts[:]

        for key, firstinfo in info_range_starts.items():
            lastinfo = info_range_ends.get(key)
            if lastinfo is None:
                warnmsgf("%s unicode range has start, but no end %r",
                         key, firstinfo)
                continue

            info = copy.copy(firstinfo)
            info.code = None
            info.name = None
            self.info_ranges.append(CharRange(firstinfo.code, lastinfo.code,
                                              info))
            self.info_range_starts.append(firstinfo.code)

    def populate_nameslist(self, download: bool=False):
        """Populate the names list by reading NamesList.txt.

        It will be downloaded if download=True is passed."""

        with self.open(UNICODE_NAMESLIST_FILE, default=_empty_file(),
                                               download=download) as src:
            blocks = []

            block = None
            subheader = None
            char = {}

            def finish_last_char():
                if 'code' in char:
                    if char['code'] in self.info_by_code:
                        code = char.pop('code')
                        info = self.info_by_code[code]

                        # UnicodeData lacks the name for many characters
                        # (e.g. Hangul Syllables), but insists on naming
                        # them using < and >
                        if info.name and not name.startswith('<'):
                            del char['name']
                        info.update(**char)

                    else:
                        info = CharInfo(registry=self, **char)
                        self.info_by_code[info.code] = info
                        self.info_by_name[info.name] = info
                        self.infos_by_name[info.name].append(info)

            for line in src:
                stripped = line.strip()

                if not stripped or stripped.startswith(';'):
                    continue

                if line.startswith('@'):
                    match = nameblock_line_rgx.search(line)
                    if match is not None:
                        start, name, end = match.groups()
                        block = UnicodeBlock(int(start, 16),
                                             int(end, 16), name)
                        i = bisect.bisect(self.unicode_blocks, block) - 1
                        if (0 <= i < len(self.unicode_blocks) and
                            self.unicode_blocks[i] == block):
                                block = self.unicode_blocks[i]
                        blocks.append(block)
                        continue

                    match = subheader_line_rgx.search(line)
                    if match is not None:
                        subheader = match.group(1)

                    continue

                match = code_name_line_rgx.search(line)
                if match is not None:
                    finish_last_char()

                    char = dict(code=int(match.group(1), 16),
                                name=match.group(2),
                                block=block, subheader=subheader)
                    continue

                match = extra_rgx.search(line)
                if match is not None:
                    symbol, text = match.groups()
                    key = NAMELIST_PREFIXES.get(symbol)
                    if key is not None:
                        char.setdefault(key, []).append(text)

            finish_last_char()

    def populate_value_aliases(self, download=False):
        with self.open(UNICODE_VALUE_ALIASES_FILE,
                       default=_empty_file(), download=download) as src:

            for line in src:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue

                values = line.split(';')

                if len(values) < 3:
                    continue

                prop, abbrev, long = (s.strip() for s in values[:3])
                self.alias_values[prop][abbrev] = long
                self.value_aliases[prop][long] = abbrev
                self.value_aliases_all[prop][long].append(abbrev)

    def populate_hangul_syllables(self, download: bool=False):
        """At some point, this function may populate Hangul syllables.

        For now, Python's unicode data is used for this, which is a much saner
        way to deal with this issue.
        """

        return

        with self.open(UNICODE_HANGUL_SYLLABLE_FILE,
                       default=_empty_file(), download=download) as src:

            for line in src:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue


def all_character_codes(noncharacters: bool=False) -> IteratorOf[int]:
    """Iterate over all unicode character codes.

    If noncharacters=True is passed, this function also includes
    non-characters. The non-characters are defined in NON_CHARACTER_CODES"""
    result = chain(range(UNICODE_START, SURROGATE_START),
                   range(SURROGATE_STOP, UNICODE_STOP))
    if noncharacters:
        return result
    return (char for char in result if char not in NON_CHARACTER_CODES)


def all_characters(*args, **kwargs) -> IteratorOf[str]:
    """Iterate over all unicode characters.

    If noncharacters=True is passed, this function also includes
    non-characters. The non-characters are defined in NON_CHARACTER_CODES"""
    return map(chr, all_character_codes(*args, **kwargs))




# -*- coding: utf-8 -*-
#
#    TypeAtlas Block Math
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

"""A library that allows one to perform basic operations on discrete blocks
of e.g. codepoints defined by first and last number. They are simple,
defined by (lists of) two-element tuples. For more complex
operations, look in the rangemath module."""

from itertools import chain
from collections import namedtuple
from operator import itemgetter
from typeatlas.util import generic_type
import bisect


Union = generic_type('Union')
Optional = generic_type('Optional')
TupleOf = generic_type('Tuple')
IterableOf = generic_type('Iterable')
IteratorOf = generic_type('Iterator')


class _BlockLikeMeta(type):

    def __instancecheck__(cls, instance):
        return (isinstance(instance, tuple) and
                len(instance) >= 2 and (isinstance(x, int) for x in instance[:2]))

    def __subclasscheck__(cls, subclass):
        return issubclass(subclass, tuple)


class BlockLike(metaclass=_BlockLikeMeta):
    """A tuple of at least two integer arguments, specifying start and
    end."""

    def __init__(self):
        raise TypeError

Block = namedtuple('Block', 'start end')


_UNSPECIFIED = object()


def iterblock(block: BlockLike) -> IteratorOf[int]:
    """Iterate over the indexes (e.g. characters) inside any block
    described by a (start, end) inclusive tuple"""
    return range(block[0], block[1] + 1)


def iterblocks(blocks: IterableOf[BlockLike]) -> IteratorOf[int]:
    """Iterate over the indexes (e.g. characters) inside an iterable
    of blocks  described by a (start, end) inclusive tuples."""
    return chain.from_iterable(range(block[0], block[1] + 1)
                               for block in blocks)


def blocklen(block: BlockLike) -> int:
    """Return the length of a block"""
    return block[1] - block[0] + 1


def blockslen(blocks: IterableOf[BlockLike]) -> int:
    """Return the length of a block"""
    return sum(block[1] - block[0] + 1 for block in blocks)


def toblocks_inorder(values: IterableOf[int]) -> IteratorOf[BlockLike]:
    """Like toblocks, but does not sort. This can be an optimisation if
    the values are already sorted, or if you're doing something unusual
    (e.g. finding the continuous blocks in an unordered sequence)"""

    start = None
    end = None

    for i in values:
        if end != i - 1:
            if start is not None:
                yield Block(start, end)
            start = i
        end = i

    if start is not None:
        yield Block(start, end)


def have_intersection(a: BlockLike, b: BlockLike) -> bool:
    """Return True if the two blocks have an intersection."""

    # Verify both implementations work

    astart = a[0]
    aend = a[1]

    bstart = b[0]
    bend = b[1]

    # No need to check if bstart <= aend <= bend, as then either
    # bstart <= astart <= bend or astart <= bstart <= aend.
    return (astart <= bstart <= aend or astart <= bend <= aend or 
            bstart <= astart <= bend)


def intersection(a: BlockLike, b: BlockLike,
                 default=_UNSPECIFIED) -> Optional[BlockLike]:
    """Return the intersection of the two blocks."""

    start = max(a[0], b[0])
    end = min(a[1], b[1])

    if start <= end:
        return Block(start, end)

    if default is _UNSPECIFIED:
        raise ValueError("blocks have no intersection")
    return default


def toblocks(values: IterableOf[int]) -> IteratorOf[BlockLike]:
    """Turn an unsorted sequence of integers into a sequence of 
    Block named tuples describing the start and end (inclusive)
    of the continuous numbers."""
    return toblocks_inorder(sorted(values))


OverlapTuple = TupleOf[BlockLike, BlockLike, BlockLike]

#def overlapping_blocks(aseq: IterableOf[int],
#                       bseq: IterableOf[int]) -> IteratorOf[OverlapTuple]:
#    """Return an iterable of overlapping blocks from the aseq and bseq,
#    as a tuple of (a, b, overlap), where a is a block from aseq, b is 
#    a block from bseq, and overlap is a block with their overlap."""
#
#    aseq = sorted(aseq)
#    astarts = list(map(itemgetter(0), aseq))
#    aends = list(map(itemgetter(1), aseq))
#    bseq = sorted(bseq)
#
#    for b in bseq:
#        bstart = b[0]
#        bend = b[1]
#
#        #i = bisect.bisect_right(astarts, bstart) - 1
#        #j = bisect.bisect_left(aends, bend) + 1
#        
#        i = bisect.bisect_left(bends, astart)
#        j = bisect.bisect_right(bstarts, aend)
#
#        for a in aseq[i:j]:
#            overlap = intersection(a, b)
#            if overlap is not None:
#                yield a, b, overlap


def overlapping_blocks(aseq: IterableOf[BlockLike],
                       bseq: IterableOf[BlockLike],
                       asymmetric: bool=False
                       ) -> IteratorOf[Union[BlockLike, OverlapTuple]]:

    """Return an iterable of overlapping blocks from the aseq and bseq,
    as a tuple of (a, b, overlap), where a is a block from aseq, b is 
    a block from bseq, and overlap is a block with their overlap.
    
    If asymmetric is True, only return the blocks from aseq that 
    overlap with some block from bseq. In other words, aseq is a haystack, 
    and bseq is a sequence of needles to locate the blocks from the 
    haystack."""

    aseq = sorted(aseq)
    bseq = sorted(bseq)
    bstarts = list(map(itemgetter(0), bseq))
    bends = list(map(itemgetter(1), bseq))

    for a in aseq:
        astart = a[0]
        aend = a[1]

        ## These are very wrong, but I'm keeping to figure out why I thought
        ## this would ever make sense.
        #i = bisect.bisect_right(bstarts, astart) - 1
        #j = bisect.bisect_left(bends, aend) + 1

        i = bisect.bisect_left(bends, astart)
        j = bisect.bisect_right(bstarts, aend)

        for b in bseq[i:j]:
            overlap = intersection(a, b, None)
            if overlap is not None:
                if asymmetric:
                    yield a
                    break
                yield a, b, overlap


def intersect_many(aseq: IterableOf[BlockLike],
                   bseq: IterableOf[BlockLike]) -> IteratorOf[BlockLike]:
    """Return the intersection of the two sequences of blocks.

    This is used to unit test overlap() and overlapping_blocks() in
    one go. For this complex an operation, rangemath may be more useful.
    """
    yield from map(itemgetter(2), overlapping_blocks(aseq, bseq))


def union(*block_sequences: IterableOf[BlockLike],
          merge_adjacent: bool=True) -> IteratorOf[BlockLike]:
    """Given a few block sequences, yield a single block iterable
    with the union of all blocks. You can also pass a single
    sequence of overlapping blocks, and the function will still work.


    Adjacent blocks will be merged, unless you pass merge_adjacent=True.
    """

    blocks = sorted(chain.from_iterable(block_sequences))

    adjoff = 1 if merge_adjacent else 0

    i = 0

    while i < len(blocks):
        start = blocks[i][0]
        end = blocks[i][1]

        # Until the next block overlaps, merge with the current one, and
        # skip it in the main loop.
        for j in range(i + 1, len(blocks)):

            # To return the promised result, we only need to break when the
            # start of the next is larger than the end, but if also keep
            # going when the start is right next to the end, we'd also merge
            # adjacent blocks.
            if blocks[j][0] > end + adjoff:
                break
            end = max(end, blocks[j][1])
            i += 1

        i += 1

        yield Block(start, end)



# -*- coding: utf-8 -*-
#
#    TypeAtlas About Dialogs
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#

"""The about dialogs, acknowledging everyone providing fonts on
your system."""

import typeatlas.proginfo as proginfo
import typeatlas.osinfo as osinfo
from typeatlas.compat import QtCore, QtGui, QtWidgets, Qt, Slot
from typeatlas.uitools import getIcon, getImage, Toolbox, iconSize
from typeatlas.uitools import generalSize, generalWidth, generalHeight
from typeatlas.langutil import _, N_

import os
import platform
from collections.abc import Iterator


def w(refWidth: int) -> int:
    """Get widgth"""
    return generalWidth(refSize=refWidth)

def h(refHeight: int) -> int:
    """Get height"""
    return generalHeight(refSize=refHeight)


px = w
py = h


class TypeAtlasAbout(QtWidgets.QDialog):

    def __init__(self, parent: QtWidgets.QWidget=None):
        super(TypeAtlasAbout, self).__init__(parent)
        self.setWindowTitle(_("About %s") % (_(proginfo.PROGRAM_NAME), ))

        self.image = QtWidgets.QLabel(self)
        pixmap = getImage('splash').pixmap(w(640), h(640))
        self.image.setPixmap(pixmap)
        self.resize(pixmap.size())

        text = [_(proginfo.PROGRAM_NAME) + ' ' + proginfo.VERSION_HUMAN_SHORT]


        self.text = QtWidgets.QLabel(self)
        self.text.setTextFormat(Qt.PlainText)
        self.text.setText('\n'.join(text))
        self.text.move(px(10), py(170))

        self.buttons = buttons = QtWidgets.QDialogButtonBox(self)
        buttons.addButton(buttons.Ok)
        buttons.accepted.connect(self.accept)

        self.buttons.move(px(12), py(330))


class FreeTypeAbout(QtWidgets.QDialog):

    def __init__(self, parent: QtWidgets.QWidget=None, version: str=None):
        super(FreeTypeAbout, self).__init__(parent)
        self.setWindowTitle(_("About FreeType"))

        if version is None:
            version = osinfo.freetype_version()

        self.image = QtWidgets.QLabel(self)
        self.image.setPixmap(getImage('freetype-logo', '.png').pixmap(584, 338))
        self.resize(w(584), h(338))

        self.text = QtWidgets.QLabel(self)
        self.text.setTextFormat(Qt.PlainText)
        text = [_("Fonts rendering provided by FreeType %s") % (version, ),
                "",
                _("FreeType is a cross-platform multi-format library "
                  "for font rasterization.")]

        self.text.setText('\n'.join(text))
        self.text.move(px(12), py(220))

        self.buttons = buttons = QtWidgets.QDialogButtonBox(self)
        buttons.addButton(buttons.Ok)
        buttons.accepted.connect(self.accept)

        self.buttons.move(px(12), py(300))


class FontConfigAbout(QtWidgets.QDialog):

    def __init__(self, parent: QtWidgets.QWidget=None, version: str=None):
        super(FontConfigAbout, self).__init__(parent)
        self.setWindowTitle(_("About fontconfig"))

        if version is None:
            from typeatlas import fontlist
            version = fontlist.FontFinder().fontconfig_version()

        self.image = QtWidgets.QLabel(self)
        self.image.setPixmap(getImage('freedesktop-logo').pixmap(586, 91))
        self.resize(w(596), h(240))

        self.text = QtWidgets.QLabel(self)
        self.text.setTextFormat(Qt.PlainText)
        text = [_("TypeAtlas is powered by the font database of fontconfig %s")
                                    % (version, ),
                "",
                _("fontconfig is freedesktop.org's library for font "
                  "discovery, substitution and configuration.")]

        self.text.setText('\n'.join(text))
        self.text.move(px(12), py(120))

        self.buttons = buttons = QtWidgets.QDialogButtonBox(self)
        buttons.addButton(buttons.Ok)
        buttons.accepted.connect(self.accept)

        self.buttons.move(px(12), py(190))



class PythonAbout(QtWidgets.QDialog):

    def __init__(self, parent: QtWidgets.QWidget=None):
        super(PythonAbout, self).__init__(parent)
        self.setWindowTitle(_("About Pythonⓡ"))

        layout = QtWidgets.QVBoxLayout()

        self.icon = QtWidgets.QLabel()
        self.icon.setPixmap(getImage('python-powered-w').pixmap(generalSize(refSize=256)))
        text = [
                _('This program is powered by <b>Python®</b> %s')
                            % (platform.python_version(), ),
                "",
                _("Python is a cross-platform high-level programming language"),
                "",
                _("Currently using %s implementation")
                            % (platform.python_implementation(), ),
                "",
        ]

        if platform.python_implementation() == 'CPython':
            text.extend([
                _('The Python language is distributed under the terms of the '
                  '<a href="%s">PSF License Agreement.</a>')  % (
                            ('https://docs.python.org/3/license.html', )),

            ])

        text.extend([
                _('“Python” and the Python logos are trademarks or registered '
                  'trademarks of the Python Software Foundation, used in '
                  'TypeAtlas under <a href="%s">PSF Trademark Usage Policy</a> '
                  'as published by the Foundation.') % (
                            ('https://www.python.org/psf/trademarks/', )),
        ])

        self.text = QtWidgets.QLabel()
        self.text.setTextFormat(Qt.RichText)
        self.text.setWordWrap(True)
        self.text.setText('<br>\n'.join(text))

        self.buttons = buttons = QtWidgets.QDialogButtonBox()
        buttons.addButton(buttons.Ok)
        buttons.accepted.connect(self.accept)

        layout.addWidget(self.icon)
        layout.addWidget(self.text)
        layout.addWidget(self.buttons)
        layout.setAlignment(Qt.AlignHCenter)

        self.setLayout(layout)


class SystemAbout(QtWidgets.QDialog):

    def __init__(self, parent: QtWidgets.QWidget=None):
        super(SystemAbout, self).__init__(parent)
        self.setWindowTitle(_("About operating system"))

        layout = QtWidgets.QHBoxLayout()

        self.icon = QtWidgets.QLabel()
        if osinfo.oslogo.icon:
            pixmap = getIcon(osinfo.oslogo.icon).pixmap(iconSize(refSize=64))
            self.icon.setPixmap(pixmap)
        elif osinfo.oslogo.symbol:
            self.icon.setText(osinfo.oslogo.symbol)
            if osinfo.oslogo.font:
                self.font = QtGui.QFont(osinfo.oslogo.font)
                self.font.setPointSize(48)
                self.icon.setFont(self.font)

        text = [platform.node()]
        text.append(' '.join(platform.system_alias(platform.system(),
                                                   platform.release(),
                                                   platform.version())))
        distname, version, distid = platform.dist()
        if distname:
            text.append(' '.join([distname.capitalize(), version, distid]))

        text.append(_('System: %s') % (platform.machine(), ))

        self.text = QtWidgets.QLabel()
        self.text.setTextFormat(Qt.PlainText)
        self.text.setText('\n'.join(text))

        self.buttons = buttons = QtWidgets.QDialogButtonBox()
        buttons.addButton(buttons.Ok)
        buttons.accepted.connect(self.accept)

        layout.addWidget(self.icon)
        layout.addWidget(self.text)
        layout.addWidget(self.buttons)

        self.setLayout(layout)


class AboutDialogs(Toolbox):

    def __init__(self, mainWindow: QtWidgets.QWidget, *args, **kwargs):
        self.mainWindow = mainWindow
        self.freeTypeVersion = osinfo.freetype_version()
        self.fontConfigVersion = mainWindow.finder.fontconfig_version()
        super(AboutDialogs, self).__init__(*args, **kwargs)

    @property
    def actionParent(self) -> QtCore.QObject:
        return self.mainWindow

    def actionDefinitions(self) -> Iterator:
        if self.freeTypeVersion:
            yield self.aboutFreeType, _("About FreeType"),
        if self.fontConfigVersion:
            yield self.aboutFontConfig, _("About fontconfig")

        yield self.action(self.aboutQt, _("About Qt®"), icon='qtlogo')
        yield self.action(self.aboutPython, _("About Python®"),
                          icon='applications-python')
        yield self.action(self.aboutSystem, _("About operating system"),
                          icon='distributor-logo')
        yield self.separator()
        yield self.action(self.about,
                          _("About %s") % (_(proginfo.PROGRAM_NAME), ),
                          icon='help-about')

    @Slot(bool)
    def about(self, checked=None):
        """Display the dialog showing the about for TypeAtlas."""
        dialog = TypeAtlasAbout(self.mainWindow)
        dialog.exec_()

    @Slot(bool)
    def aboutFreeType(self, checked=None):
        """Display the dialog showing the about for FreeType."""
        dialog = FreeTypeAbout(self.mainWindow, self.freeTypeVersion)
        dialog.exec_()

    @Slot(bool)
    def aboutFontConfig(self, checked=None):
        """Display the dialog showing the about for fontconfig."""
        dialog = FontConfigAbout(self.mainWindow, self.fontConfigVersion)
        dialog.exec_()

    @Slot(bool)
    def aboutSystem(self, checked=None):
        """Display the dialog showing the about for the system."""
        dialog = SystemAbout(self.mainWindow)
        dialog.exec_()

    @Slot(bool)
    def aboutPython(self, checked=None):
        """Display the dialog showing the about for Python."""
        dialog = PythonAbout(self.mainWindow)
        dialog.exec_()

    @Slot(bool)
    def aboutQt(self, checked=None):
        """Display the dialog showing the about for Qt."""
        QtWidgets.QMessageBox.aboutQt(self.mainWindow)




# -*- coding: utf-8 -*-
#
#    TypeAtlas Operating System Information
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

# License note:
#
# If the compat module is loaded, this module will use Qt and PyQt or PySide
# to acquire the storage drive list, the mounted volume list, the icon theme
# search paths and the current theme name. At the time of writing Qt and PySide
# were licensed under the GNU LGPL, and PyQt was under the GNU GPL.
#
# Be mindful of that. If in doubt, the Qt code can be stripped by searching
# for qtcompat

"""Module with information about the operating system, such as supported font
types, icon finder, drive finder, etc."""

from typeatlas.foreign import appdirs
from collections import namedtuple
import re
import os
import os.path
import sys
import platform
import traceback
import configparser
from operator import attrgetter
from distutils.version import LooseVersion as version_key
#from packaging.version import parse as version_key

from typeatlas.util import warnmsgf, generic_type
from typeatlas.langutil import _, N_
from typeatlas import proginfo

IteratorOf = generic_type('Iterator')
IterableOf = generic_type('Iterable')
SequenceOf = generic_type('Sequence')


FREETYPE_COLR_VERSION_MIN = '2.10.0'

WINDOWS_COLR_VERSION_MIN = '8.1'
WINDOWS_SVG_VERSION_MIN = '10'

# FIXME: I see conflicting reports that these work in macOS 10.14, with
#        COLR/CPAL fonts as early as 10.13.
# Sources: https://yoksel.github.io/color-fonts-demo/
#          https://bugzilla.mozilla.org/show_bug.cgi?id=1425812
#          https://en.wikipedia.org/w/index.php?title=OpenType&oldid=1008183155#Color
DARWIN_COLR_VERSION_MIN = '17.0.0'
DARWIN_SVG_VERSION_MIN = '18.0.0'


OsLogo = namedtuple('OsLogo', 'icon font symbol')

LINUX_PSEUDO_FILESYSTEM = set([
    'anon_inodefs', 'autofs', 'bdev', 'binfmt_misc', 'cgroup', 'cgroup2',
    'configfs', 'cpuset', 'debugfs', 'devfs', 'devpts', 'devtmpfs', 'dlmfs',
    'efivarfs', 'fuse.gvfsd-fuse', 'fuse.gvfs-fuse-daemon', 'fusectl',
    'hugetlbfs', 'mqueue', 'nfsd', 'none', 'overlay', 'pipefs', 'proc',
    'pstore', 'ramfs', 'rootfs', 'rpc_pipefs', 'securityfs', 'sockfs',
    'spufs', 'sysfs', 'tmpfs'
])

# Optical filesystems on Linux and macOS, including the response of psutil
# on Windows.
OPTICAL_FILESYSTEM = set(['udf', 'iso9660', 'cdrom', 'cd9660', 'cdfs'])
NETWORK_FILESYSTEM = set(['remote', 'nfs', 'nfs3', 'nfs4', 'cifs', 'smbfs',
                          'ncpfs', '9p'])

# psutil on Windows only
REMOVABLE_FILESYSTEMS = set(['removable'])


is_windows = False
is_posix = False
is_linux = False
is_darwin = False

if sys.platform.startswith('linux'):
    if platform.dist()[0].lower() == 'ubuntu':
        oslogo = OsLogo(None, 'Ubuntu', u'\uE0FF')
    #oslogo = OsLogo('distributor-logo', None, None)
    oslogo = OsLogo(None, 'Linux Libertine O', u'\uE000')
    is_posix = True
    is_linux = True
elif sys.platform == 'darwin':
    oslogo = OsLogo(None, None, u'\uF8FF')
    is_posix = True
    is_darwin = True
elif sys.platform == 'win32':
    oslogo = OsLogo(None, 'Wingdings', u'\uF0FF')
    is_windows = True
elif os.name == 'posix':
    is_posix = True
    oslogo = OsLogo('distributor-logo', None, None)
else:
    oslogo = OsLogo(None, None, None)


_UNSPECIFIED = object()

_freetype_version = _UNSPECIFIED


def freetype_version() -> str:
    """Get the version of freetype currently installed on the system."""

    global _freetype_version
    if _freetype_version is not _UNSPECIFIED:
        return _freetype_version

    try:
        import ctypes
    except ImportError:
        _freetype_version = None
        return _freetype_version

    try:
        freetype = ctypes.CDLL('libfreetype.so.6')
    except (TypeError, OSError):
        _freetype_version = None
        return _freetype_version

    ft = ctypes.pointer(ctypes.c_voidp(1))

    major = ctypes.pointer(ctypes.c_int(1))
    minor = ctypes.pointer(ctypes.c_int(1))
    patch = ctypes.pointer(ctypes.c_int(1))

    if freetype.FT_Init_FreeType(ft) != 0:
        _freetype_version = None
        return _freetype_version

    try:
        freetype.FT_Library_Version(ft.contents, major, minor, patch)

        _freetype_version = '%d.%d.%d' % (major.contents.value,
                                          minor.contents.value,
                                          patch.contents.value)

    finally:
        freetype.FT_Done_FreeType(ft.contents)

    return _freetype_version


_layered_fonts_supported = None


def layered_fonts_supported() -> bool:
    """Return True if the platform supports CPAL/COLR layered fonts.
    The value may be False on platforms that do."""

    global _layered_fonts_supported
    if _layered_fonts_supported is not None:
        return _layered_fonts_supported

    if is_windows:
        _layered_fonts_supported = (version_key(platform.version()) >=
                                    version_key(WINDOWS_COLR_VERSION_MIN))
    elif is_darwin:
        _layered_fonts_supported = (version_key(platform.release()) >=
                                    version_key(DARWIN_COLR_VERSION_MIN))
    elif is_posix:
        # FIXME: Check features?
        version = freetype_version()
        _layered_fonts_supported = (version_key(version) >=
                                    version_key(FREETYPE_COLR_VERSION_MIN))
    else:
        _layered_fonts_supported = False

    return _layered_fonts_supported


_svg_fonts_supported = None


def svg_fonts_supported() -> bool:
    """Return True if the platform supports SVG fonts. The value may be
    False on platforms that do."""

    global _svg_fonts_supported
    if _svg_fonts_supported is not None:
        return _svg_fonts_supported

    if is_windows:
        _svg_fonts_supported = (version_key(platform.version()) >=
                                version_key(WINDOWS_SVG_VERSION_MIN))
    elif is_darwin:
        _svg_fonts_supported = (version_key(platform.release()) >=
                                version_key(DARWIN_SVG_VERSION_MIN))
    elif is_posix:
        # FreeType does not yet support SVG fonts
        _svg_fonts_supported = False
    else:
        _svg_fonts_supported = False

    return _svg_fonts_supported


class StorageLocation(object):

    def __init__(self, path: str, label: str=None, icon: str='folder'):
        if label is None:
            label = path

        self.path = path
        self.label = label
        self.icon = icon


class StorageLocations(object):

    def __init__(self, *args, **kwargs):
        super(StorageLocations, self).__init__(*args, **kwargs)

        try:
            import psutil
        except ImportError:
            self.psutil = None
        else:
            self.psutil = psutil

        self.win32api = None

        if is_windows:
            try:
                import win32api
            except ImportError:
                win32api = None
            self.win32api = win32api

        self.qtcompat = sys.modules.get('typeatlas.compat')

    def drives(self) -> IteratorOf[StorageLocation]:
        """Yield the disk drives of the computer, e.g. flash drives."""
        if self.win32api is not None:
            for drive in self.win32api.GetLogicalDriveStrings().split('\0'):
                if not drive:
                    continue
                path = os.path.normpath(drive + ':\\')
                yield StorageLocation(path, _('Drive %s') % (path, ),
                                      icon='drive-harddisk')
            return

        if self.qtcompat is not None:
            for fi in self.qtcompat.QtCore.QDir.drives():
                path = os.path.normpath(str(fi.absoluteFilePath()))
                yield StorageLocation(path, _('Drive %s') % (path, ),
                                      icon='drive-harddisk')


    def locations(self) -> IteratorOf[StorageLocation]:
        """Yield the special storage locations on the computer, e.g. homedir."""
        seen = set()
        seen_devs = set()

        homedir = os.path.expanduser('~')
        seen.add(homedir)

        if os.path.exists(homedir):
            yield StorageLocation(homedir, _('Home directory %s') % (homedir, ),
                                  icon='user-home')

        if is_posix and '/' not in seen:
            seen.add('/')
            yield StorageLocation('/', _('Root /'), icon='system')

        if self.psutil is not None:
            for disk in sorted(self.psutil.disk_partitions(all=True),
                               key=attrgetter('mountpoint')):

                if disk.mountpoint in seen:
                    seen_devs.add(disk.device)
                    continue
                if disk.device in seen_devs:
                    seen.add(disk.mountpoint)
                    continue

                seen.add(disk.mountpoint)
                seen_devs.add(disk.device)

                fstype = disk.fstype.lower()

                if fstype in LINUX_PSEUDO_FILESYSTEM:
                    continue

                if not os.access(disk.mountpoint, os.X_OK):
                    continue

                options = disk.opts.split(',')

                if fstype in OPTICAL_FILESYSTEM or 'cdrom' in options:
                    label = _('Optical drive %s') % (disk.mountpoint, )
                    icon = 'drive-optical'

                elif fstype in NETWORK_FILESYSTEM or 'remote' in options:
                    label = _('Remote folder %s') % (disk.mountpoint, )
                    icon = 'folder-remote'

                elif (fstype in REMOVABLE_FILESYSTEMS or
                      (is_linux and '/media/' in disk.mountpoint) or
                      'removable' in options):

                    label = _('Removable drive %s') % (disk.mountpoint, )
                    icon = 'drive-removable-media'

                else:
                    if is_windows:
                        label = _('Drive %s') % (disk.mountpoint, )
                    else:
                        label = _('Mountpoint %s') % (disk.mountpoint, )
                    icon = 'drive-harddisk'

                yield StorageLocation(disk.mountpoint, label, icon )

        elif is_windows:
            for drive in self.drives():
                if drive.path in seen:
                    continue
                yield drive

        elif self.qtcompat is not None:
            for storage in self.qtcompat.QtCore.QStorageInfo.mountedVolumes():
                if storage.rootPath() in seen:
                    continue
                yield StorageLocation(storage.rootPath(), storage.displayName(),
                                      icon='folder')


_icon_search_paths = None


ThemeIconInfo = namedtuple('ThemeIconInfo', 'name context')


def find_icons(theme_name: str=None,
               search_paths: SequenceOf[str]=None,
               inherit: bool=False,
               parse_theme: bool=True, strict: bool=False,
               skip: IterableOf[str]=None,
               include_ours: bool=True,
               hide_duplicates: bool=True) -> IteratorOf[ThemeIconInfo]:
    """Find the icons from a given theme in the given search paths,
    and yield ThemeIconInfo named tuples.

    The theme name and paths are guessed if not passed. If
    Qt is loaded, it is used to acquire them.

    If inherit is True, theme inheritance is observed to an extend,
    but the result is slower. If parse_theme is False, no attempt
    is made to parse the theme, resulting in incorrect context.
    A skip iterable can be passed to skip certain icons, this is
    mainly used so that this function can recurse.

    If strict=True is passed, configparser errors will be propagated.
    """

    global _icon_search_paths

    if theme_name is None or search_paths is None:
        qtcompat = sys.modules.get('typeatlas.compat')
    else:
        qtcompat = None

    if theme_name is None:
        if qtcompat is not None:
            theme_name = qtcompat.QtGui.QIcon.themeName()
        else:
            theme_name = 'hicolor'

    if search_paths is None:
        search_paths = []

        if _icon_search_paths is not None:
            search_paths = _icon_search_paths

        elif qtcompat is not None:
            search_paths = qtcompat.QtGui.QIcon.themeSearchPaths()

        elif os.name == 'posix':
            seen_paths = set()

            for icon_dir in [
                        appdirs.user_data_dir('icons'),
                        appdirs.site_data_dir('icons'),
                        '/usr/share/icons',
                        '/usr/local/share/icons',
                        os.path.expanduser('~/.icons')]:

                if (icon_dir and icon_dir not in seen_paths and
                    os.path.isdir(icon_dir)):

                        search_paths.append(icon_dir)
                        seen_paths.add(icon_dir)

            _icon_search_paths = tuple(search_paths)

    seen = set(skip or ())

    if include_ours:
        try:
            files = os.listdir(proginfo.PROGRAM_ICON_DIR)
        except OSError as exc:
            warnmsgf("Could not get our icons from %s: %s: %s",
                     proginfo.PROGRAM_ICON_DIR, type(exc).__name__, exc)
        else:
            for filename in files:
                basename, ext = os.path.splitext(filename)
                filepath = os.path.join(proginfo.PROGRAM_ICON_DIR, filename)
                if os.path.isfile(filepath) and ext:
                    yield ThemeIconInfo(basename, proginfo.PROGRAM_SHORT_NAME)


    for root in search_paths:
        theme_path = os.path.join(root, theme_name)
        if not os.path.exists(theme_path):
            continue

        inherited_themes = []
        directory_context = {}
        has_index_theme = False

        # Try to extract icon contexts and inherited themes from
        # the index.theme file
        theme_index = os.path.join(theme_path, 'index.theme')
        if parse_theme and os.path.exists(theme_index):
            parser = configparser.RawConfigParser(strict=False)
            try:
                parser.read_file(open(theme_index))
                for section in parser.sections():
                    if section == 'Icon Theme':
                        if 'inherits' in parser.options(section):
                            inherited_themes.extend(
                                    parser.get(section, 'inherits')
                                          .split(','))
                        continue
                    if 'context' not in parser.options(section):
                        continue

                    reldir = os.path.normpath(section)
                    directory_context[reldir] = parser.get(section, 'context')
                has_index_theme = True

            except (configparser.Error, OSError, UnicodeError) as exc:
                if strict:
                    raise
                warnmsgf("Could not read theme index %s: %s: %s",
                         theme_index, type(exc).__name__, exc)
                traceback.print_exc()

        for dirname, dirnames, filenames in os.walk(theme_path):
            reldir = os.path.normpath(os.path.relpath(dirname, theme_path))
            context = directory_context.get(reldir)

            dirnames.sort(key=lambda s: s.casefold())
            filenames.sort(key=lambda s: s.casefold())

            # Try to guess from directory name, somehow
            if context is None:
                # If this didn't path didn't exist in index.theme, ignore.
                if has_index_theme:
                    continue

                components = [comp
                              for comp in reldir.split(os.sep)
                              if not re.match('[0-9]+(?:x[0-9]+)?$', comp)]
                if components:
                    context = os.sep.join(components)
                else:
                    context = 'misc'
                directory_context[reldir] = context

            for filename in filenames:
                if hide_duplicates:
                    filepath = os.path.join(dirname, filename)
                    if os.path.islink(filepath):
                        target = os.path.normpath(os.path.join(
                                    dirname, os.readlink(filepath)))
                        if (os.path.normpath(dirname) == os.path.normpath(
                                os.path.commonpath([target, dirname]))):
                            continue

                basename, ext = os.path.splitext(filename)
                if basename in seen:
                    continue

                seen.add(basename)
                yield ThemeIconInfo(basename, context)

        if inherit and inherited_themes:
            for inherited_theme in inherited_themes:
                for icon_info in find_icons(theme_name, search_paths,
                                            inherit=False, # FIXME ???
                                            parse_theme=parse_theme,
                                            strict=strict, skip=seen,
                                            hide_duplicates=hide_duplicates):
                    seen.add(icon_info.name)
                    yield icon_info



# -*- coding: utf-8 -*-
#
#    TypeAtlas CLI Finder Functions
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas command-line)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import division

import re
import os
import sys
import glob
import fnmatch
import argparse
import operator
import shutil

from itertools import groupby, chain
from operator import attrgetter
from typeatlas import fontlist, filtering
from typeatlas.langutil import _, N_, textlang
from typeatlas.util import isatty, tty_clear


FAMILY_PRINT_FORMAT = '{family}\n'
STYLE_PRINT_FORMAT = '{family} {style}\n'
FILE_PRINT_FORMAT = '{file}\n'


AUTO = object()


find_argparser = argparse.ArgumentParser(_('TypeAtlas font finder'))

matcher_dests = set()
matcher_factories = {}


comparisons = {
    # Readable operators
    '=': operator.eq,
    '!=': operator.ne,
    '!': operator.ne,
    '>': operator.gt,
    '<': operator.lt,
    '>=': operator.ge,
    '<=': operator.le,

    # Shell-safe operators, similar to find
    '++': operator.gt,
    '--': operator.lt,
    '+': operator.ge,
    '-': operator.le,

}


def split_comparison(arg):

    for op, func in comparisons.items():
        if arg.startswith(op):
            return func, arg[len(op):]

    return None, arg


opening_bracket = frozenset('{([')
closing_bracket = frozenset('})]')
matching_bracket = dict(['{}', '()', '[]', '}{', ')(', ']['])


class ParseError(ValueError):
    pass


def parse_expression(args):

    """Parse a font matching expression using the defined predicates"""

    return _parse_expression_lowlevel(args)[1]


def _parse_expression_lowlevel(args, i=0, expect_bracket=None, single=False):

    """Parse a font matching expression using the defined predicates
    that starts at the given index. If a paren is open, pass the closing
    paren in expect_bracket."""

    CONJ = object()

    got_paren = False

    predicate_group = []
    predicate_groups = [predicate_group]

    while i < len(args):
        arg = args[i]

        if arg in closing_bracket:
            if arg != expect_bracket:
                raise ParseError(_("Unexpected %r") % (arg, ))
            got_paren = True
            i += 1
            break

        elif arg in opening_bracket:
            i, matcher = _parse_expression_lowlevel(args, i + 1,
                                                    matching_bracket[arg])
            predicate_group.append(matcher)
            continue

        elif arg == 'or':
            predicate_group = []
            predicate_groups.append(predicate_group)

        elif arg == 'and':
            pass

        elif arg == 'not':
            i, matcher = _parse_expression_lowlevel(args, i + 1, single=True)
            predicate_group.append(Negation(matcher))
            continue

        else:
            try:
                matcher_cls = matcher_factories[arg]

            except KeyError:
                raise ParseError(_("Invalid predicate %r") % (arg, ))

            else:
                if len(args) < i + matcher_cls.nargs + 1:
                    raise ParseError("Missing arguments")
                matcher_args = args[i + 1:i + matcher_cls.nargs + 1]
                i += matcher_cls.nargs
                predicate_group.append(matcher_cls(*matcher_args))

        i += 1

        if single:
            break

    if expect_bracket and not got_paren:
        raise ParseError(_("Missing %s") % (expect_bracket, ))

    if not all(predicate_groups):
        raise ParseError(_("Empty expression"))

    return i, Disjunction.make(map(Conjunction.make, predicate_groups))



def add_matcher(*options, disjunctive_group=None, nargs=None, **kwargs):

    """Register a matcher as an option and an expression predicate."""

    def decorator(cls):
        if nargs == 0:
            const = cls()
            action = 'append_const'
        else:
            const = None
            action = 'append'
            kwargs.update(type=cls)
            if nargs is not None:
                kwargs.update(nargs=nargs)

        arg = find_argparser.add_argument(*options,
                                          action=action,
                                          const=const,
                                          **kwargs)
        cls.argspec = arg
        cls.disjunctive_group = disjunctive_group
        cls.nargs = 1
        if arg.nargs is not None:
            if not isinstance(arg.nargs, int):
                raise TypeError("nargs must be integer")
            cls.nargs = arg.nargs

        matcher_dests.add(arg.dest)
        matcher_factories[arg.dest] = cls
        return cls

    return decorator


def get_matchers(args):

    expression = args.expression
    if expression:
        yield parse_expression(expression)

    for dest in matcher_dests:
        matchers = getattr(args, dest) or ()
        for matcher in matchers:
            matcher.set_args(args)
            yield matcher


class Matcher(object):
    """Base for classes that match/select fonts."""

    disjunctive_group = None
    uses_extended = False
    slow = False
    args = None

    def __init__(self, arg):
        self.arg = arg

    def set_args(self, args):
        self.args = args

    def match(self, font):
        return True

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.arg)


class Conjunction(Matcher):

    """A conjunction ("and") of matchers"""

    def __init__(self, matchers):
        self.matchers = matchers

    @classmethod
    def make(cls, matchers):
        matchers = list(matchers)
        if len(matchers) == 1:
            return matchers[0]
        return cls(matchers)

    def set_args(self, args):
        for matcher in self.matchers:
            matcher.set_args(args)
        self.slow = any(matcher.slow for matcher in self.matchers)
        self.uses_extended = any(matcher.uses_extended
                                 for matcher in self.matchers)

    def match(self, font):
        return all(matcher.match(font) for matcher in self.matchers)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.matchers)


class Disjunction(Matcher):

    """A conjunction ("or") of matchers"""

    def __init__(self, matchers):
        self.matchers = matchers

    @classmethod
    def make(cls, matchers):
        matchers = list(matchers)
        if len(matchers) == 1:
            return matchers[0]
        return cls(matchers)

    def set_args(self, args):
        for matcher in self.matchers:
            matcher.set_args(args)
        self.slow = any(matcher.slow for matcher in self.matchers)
        self.uses_extended = any(matcher.uses_extended
                                 for matcher in self.matchers)

    def match(self, font):
        return any(matcher.match(font) for matcher in self.matchers)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.matchers)


class Negation(Matcher):

    """A negation ("not") of matchers"""

    def __init__(self, matcher):
        self.matcher = matcher

    def set_args(self, args):
        self.matcher.set_args(args)
        self.slow = self.matcher.slow
        self.uses_extended = self.matcher.uses_extended

    def match(self, font):
        return not self.matcher.match(font)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.matcher)


def _make_name_matcher(cat, prefix, opt, multi, langs, extended,
                       regex, casesensitive):

    """Create a name matcher for the given category, using the given prefix
    prepended to the name, with the given short option.

    multi is True if fontlist.Font supports multiple arguments here,
    langs is True if it supports localization, extended if it is gotten from
    the extended info, regex is True if we should create regex match, and
    casesensitive is True if we want to create a case sensitive matcher.

    The lowercase category with spaces replaced with underscores is the
    attribute used to get the property.]
    """

    if regex:
        name = cat.capitalize() + 'Regex'
        help = _('Select fonts whose %s matches a regular expression') \
                    % (_(cat), )
    else:
        name = cat.capitalize() + 'Name'
        help = _('Select fonts whose %s contains a string or wildcard pattern. '
                 'Prefix with = to to get exact match, or with ~ to '
                 'match the pattern against the whole name.') \
                    % (_(cat), )

    attr = name.lower().replace(' ', '_')

    if regex:
        flags = '(?ms)' if casesensitive else '(?ims)'
    else:
        flags = '' if casesensitive else '(?i)'

    if regex:
        option = '--' + prefix + '-regex' if prefix else '--regex'
    else:
        option = '--' + prefix if prefix else '--name'

    if casesensitive:
        name += 'Case'
        help += '; ' + _('Use case sensitive matching.')
        option += '-case'

    options = (option, )
    if opt:
        options += (opt, )

    @add_matcher(*options, disjunctive_group=cat, help=help,
                           metavar=_('PATTERN'))
    class NameMatcher(Matcher):

        uses_extended = extended

        def __init__(self, pattern):
            self.arg = pattern
            if not regex:
                if pattern.startswith('='):
                    pattern = '^' + re.escape(pattern[1:]) + '$'
                elif pattern.startswith('~'):
                    pattern = fnmatch.translate(pattern)
                else:
                    pattern = fnmatch.translate('*' + pattern + '*')
            self.regex = re.compile(flags + pattern)

        def match(self, font):
            if extended:
                font = font.extended()

            attr = cat

            if multi:
                attr += '_alts'
            if langs:
                attr += '_by_lang'

            if langs:
                names = getattr(font, attr).values()
            else:
                names = getattr(font, attr)

            if multi:
                names = chain.from_iterable(names)
            else:
                names = (names, )

            return any(v is not None and self.regex.search(v)
                       for v in names)


def _make_name_matchers():
    """Create all the name matchers"""

    for cat, prefix, o, oci, orx, orxci, multi, langs, ext in [
        (N_('family'), '', '-n', '-N', '-x', '-X', True, True, False),
        (N_('style'), 'style', '-s', '-S', '', '', True, True, False),
        (N_('fullname'), 'fullname', '', '', '', '', True, True, False),
        (N_('file'), 'filename', '-f', '-F', '', '', False, False, False),
        (N_('foundry'), 'foundry', '', '', '', '', False, False, False),

        (N_('description'), 'description', '', '', '', '', False, False, True),
        (N_('copyright'), 'copyright', '', '', '', '', False, False, True),
        (N_('trademark'), 'trademark', '', '', '', '', False, False, True),
        (N_('license'), 'license', '', '', '', '', False, False, True),
        (N_('license URL'), 'license-url', '', '', '', '', False, False, True),
        (N_('designer'), 'designer', '', '', '', '', False, False, True),
        (N_('manufacturer'), 'manufacturer', '', '', '', '',
         False, False, True),
        (N_('vendor URL'), 'vendor-url', '', '', '', '', False, False, True),
        (N_('designer URL'), 'designer-url', '', '', '', '', False, False, True),
        (N_('sample text'), 'sample-text', '', '', '', '', False, False, True),
        (N_('version'), 'version', '', '', '', '', False, False, True),

    ]:

        _make_name_matcher(cat, prefix, o, multi, langs, ext, False, False)
        _make_name_matcher(cat, prefix, oci, multi, langs, ext, False, True)
        _make_name_matcher(cat, prefix, orx, multi, langs, ext, True, False)
        _make_name_matcher(cat, prefix, orxci, multi, langs, ext, True, True)


_make_name_matchers()


def get_character_matcher(characters, qt=AUTO, contains_all=True):

    """Return an unary function that returns True if a font passed to it
    contains the given characters. If you pass qt=True, the function will
    use Qt functions, if you pass qt=False it will use typeatlas. By
    default it attempts to use Qt and fallbacks to typeatlas."""

    require = all if contains_all else any

    characters = set(map(ord, characters))

    if qt is AUTO:

        qt = False

        #if 'DISPLAY' in os.environ:
        #    try:
        #        from typeatlas.compat import QtGui
        #    except ImportError:
        #        pass
        #    else:
        #        qt = hasattr(QtGui, 'QRawFont')

    if qt:
        from typeatlas.compat import QtGui
        fontdb = QtGui.QFontDatabase()

        def character_matcher(font):
            import gc
            gc.collect()
            qfont = fontdb.font(font.family, font.family, 16)
            supports_char = QtGui.QRawFont.fromFont(qfont).supportsCharacter
            if not require(supports_char(c) for c in characters):
                return False
            return True

        character_matcher.uses_extended = False

    elif fontlist.FontFinder.supported():

        def character_matcher(font):
            charrange = font.charset
            if not require((c in charrange) for c in characters):
                return False
            return True

        character_matcher.uses_extended = False

    else:
        def character_matcher(font):
            extended = font.extended()
            if extended is not None and extended.cmap:
                cmap = extended.cmap
                if not require((c in cmap) for c in characters):
                    return False
            else:
                return False

            return True

        character_matcher.uses_extended = True

    return character_matcher


@add_matcher('--weight', '-W', help=_("Font weight, e.g. 'bold', '>=120'"))
class Weight(Matcher):

    def __init__(self, arg):
        self.arg = arg
        comp, value = split_comparison(arg)
        if value == 'bold':
            comp = operator.gt
            value = (fontlist.WEIGHT_BOLD + fontlist.WEIGHT_NORMAL) / 2
        elif value == 'regular':
            comp = lambda val, valrange: valrange[0] < val <= valrange[0]
            value = ((fontlist.WEIGHT_NORMAL + fontlist.WEIGHT_LIGHT) / 2,
                     (fontlist.WEIGHT_BOLD + fontlist.WEIGHT_NORMAL) / 2)
        elif value == 'light':
            comp = operator.le
            value = (fontlist.WEIGHT_NORMAL + fontlist.WEIGHT_LIGHT) / 2

        else:
            value = float(value)

        self.comp = comp
        self.value = value

    def match(self, font):
        return self.comp(font.weight, self.value)


@add_matcher('--slant', '-I', help=_("Font slant, e.g. 'italic', '>=30'"))
class Slant(Matcher):


    def __init__(self, arg):
        self.arg = arg
        comp, value = split_comparison(arg)
        if value == 'oblique':
            comp = operator.gt
            value = (fontlist.SLANT_OBLIQUE + fontlist.SLANT_ITALIC) / 2
        elif value == 'italic':
            comp = operator.gt
            value = (fontlist.SLANT_ITALIC + fontlist.SLANT_NORMAL) / 2
        elif value == 'regular':
            comp = operator.le
            value = (fontlist.SLANT_ITALIC + fontlist.SLANT_NORMAL) / 2
        else:
            value = float(value)

        self.comp = comp
        self.value = value

    def match(self, font):
        return self.comp(font.slant, self.value)


@add_matcher('--width', help=_("Font slant, e.g. 'condensed', '<=70'"))
class Width(Matcher):


    def __init__(self, arg):
        self.arg = arg
        comp, value = split_comparison(arg)
        if value == 'condensed':
            comp = operator.le
            value = (fontlist.WIDTH_CONDENSED + fontlist.WIDTH_NORMAL) / 2
        elif value == 'regular':
            comp = operator.gt
            value = (fontlist.WIDTH_CONDENSED + fontlist.WIDTH_NORMAL) / 2
        else:
            value = float(value)

        self.comp = comp
        self.value = value

    def match(self, font):
        return self.comp(font.width, self.value)


@add_matcher('--characters', '--chars', '-c',
             metavar=_('ABCXYZ'),
             help=_('Select fonts that contain the given characters.'))
class Characters(Matcher):

    slow = True

    def set_args(self, args):
        self.args = args

        self.match = get_character_matcher(self.arg, qt=self.args.qt)
        self.uses_extended = self.match.uses_extended


@add_matcher('--any-character', '--any-char',
             metavar=_('ABCXYZ'),
             help=_('Select fonts that contain the given characters.'))
class AnyCharacter(Matcher):

    slow = True

    def set_args(self, args):
        self.args = args

        self.match = get_character_matcher(self.arg, qt=self.args.qt,
                                                     contains_all=False)
        self.uses_extended = self.match.uses_extended


@filtering.set_cmdline_processor
def add_filter_matcher(cls, *options, **kwargs):

    antioptions = [opt.replace('--', '--no-', 1)
                   for opt in options if opt.startswith('--')]

    if cls.is_group:
        return

    if cls.is_switch:
        @add_matcher(*options, nargs=0, **kwargs)
        class FilterMatcher(Matcher):
            def __init__(self):
                self.match = cls(True).accept

        @add_matcher(*antioptions, nargs=0, **kwargs)
        class FilterMatcher(Matcher):
            def __init__(self):
                self.match = cls(False).accept
        return

    if cls.is_invertible:
        @add_matcher(*options, **kwargs)
        class FilterMatcher(Matcher):
            def __init__(self, arg):
                self.match = cls.fromstring(arg, True).accept

        @add_matcher(*antioptions, **kwargs)
        class FilterMatcher(Matcher):
            def __init__(self, arg):
                self.match = cls.fromstring(arg, False).accept
        return

    @add_matcher(*options, **kwargs)
    class FilterMatcher(Matcher):
        def __init__(self, arg):
            self.match = cls.fromstring(arg).accept




find_argparser.add_argument('--qt', const=True, default=AUTO,
                            action='store_const', dest='qt',
                            help=_('Use Qt (default is to guess)'))

find_argparser.add_argument('--no-qt', const=False, default=AUTO,
                            action='store_const', dest='qt',
                            help=_('Do not use Qt (default is to guess). '
                                   'This breaks universal character checks.'))

find_argparser.add_argument('--progress', const=True, default=AUTO,
                            action='store_const', dest='progress',
                            help=_('Use progress (default is to guess)'))

find_argparser.add_argument('--no-progress', const=False, default=AUTO,
                            action='store_const', dest='progress',
                            help=_('Do not use progress (default is to guess)'))

if fontlist.FontFinder.fcmatch_supported():
    find_argparser.add_argument('--fc-match', help='Use fc-match pattern')
    find_argparser.add_argument('--prune', action='store_true',
                                help=_('Prune fc-match matches'))
else:
    find_argparser.set_defaults(fc_match=None, prune=False)


find_argparser.add_argument('--no-localization', '-C', action='store_true',
                            help=_("Do not localize font strings"))

find_argparser.add_argument('--show-styles', '-y', action='store_true',
                            help=_("Display styles"))

find_argparser.add_argument('--show-paths', '-l', action='store_true',
                            help=_("Display files"))

find_argparser.add_argument('--format', metavar=_('TEMPLATE'),
                            help=_("Use the TEMPLATE as printing format"))

find_argparser.add_argument('--output', '-o', type=argparse.FileType('w'),
                            default=sys.stdout, metavar=_("FILE"),
                            help=_("Write fonts to FILE instead of "
                                   "standard output"))

find_argparser.add_argument('--expression', '--expr', '-e',
                            metavar=_('PRED'),
                            nargs='+',
                            help=_("Match fonts using an expression. "
                                   "The expression can contain all predicates "
                                   "options, without the leading dashes, as "
                                   "well as keywords 'or', 'and' and "
                                   "brackets for grouping."))

def find_fonts(*a, **kw):

    """Command line utility to find fonts."""

    args = find_argparser.parse_args(*a, **kw)

    # Get Qt early so that matchers can use it
    qt = args.qt

    if qt is AUTO:

        qt = False

        #if 'DISPLAY' in os.environ:
        #    try:
        #        from typeatlas.compat import QtGui, QtWidgets
        #    except ImportError:
        #        pass
        #    else:
        #        QtGui
        #        qt = True

    if qt:
        from typeatlas.compat import QtWidgets
        from typeatlas import qfontlist
        from typeatlas.uitools import QtExecutor
        app = QtWidgets.QApplication([])
        w = QtWidgets.QWidget()
        finder = qfontlist.QtFontFinder(executor=QtExecutor(parent=w))

    else:
        finder = fontlist.FontFinder()

    matcherkey = lambda matcher: matcher.disjunctive_group or ''
    matchers = sorted(get_matchers(args), key=matcherkey)

    progress = args.progress

    uses_extended = False

    for matcher in matchers:
        if matcher.uses_extended:
            uses_extended = True
        if matcher.slow:
            if progress is AUTO:
                progress = True

    if progress is AUTO:
        progress = uses_extended

    if not isatty(sys.stderr):
        progress = False

    print_format = FAMILY_PRINT_FORMAT

    show_styles = False
    if args.show_styles:
        show_styles = True
        print_format = STYLE_PRINT_FORMAT

    if args.show_paths:
        show_styles = True
        print_format = FILE_PRINT_FORMAT

    if args.format is not None:
        print_format = args.format.replace('\\n', '\n')

    def print_item(item):

        fields = {}
        for prop in fontlist.properties:
            name = prop.propname
            if prop.localized and not args.no_localization:
                value = item.translate(name, textlang())
            else:
                value = getattr(item, name)

            fields[name] = value
        print(print_format.format(**fields), file=args.output, end='')

    def match_item(item):
        # This gets read every time its accessed, unless we keep reference to
        # it.
        if uses_extended:
            extended = item.extended()
            extended

        for disjunctive_group, matcher_group in groupby(matchers, matcherkey):
            require = all if not disjunctive_group else any
            if not require(matcher.match(item) for matcher in matcher_group):
                return False
        return True

    if args.fc_match:
        fonts = finder.fcmatch(args.fc_match, many=True, all=not args.prune)
    else:
        fonts = finder.fonts()

    families = list(fontlist.get_families(fonts))
    def start_check():
        if progress:
            part_done = i / len(families)
            stars = int(part_done * 15)
            spaces  = int(15 - stars)
            sys.stderr.write('[' + '*' * stars + ' ' * spaces + ']: ')
            sys.stderr.write('%3.0f%%; ' % (part_done * 100, ))
            sys.stderr.write(_('%d out of %d families processed...')
                                    % (i, len(families)))
            sys.stderr.flush()

    def end_check():
        if progress:
            sys.stderr.write(tty_clear(sys.stderr))
            sys.stderr.flush()

    for i, family in enumerate(families):
        if not show_styles:
            start_check()
            is_match = any(match_item(style) for style in family.styles)
            end_check()

            if is_match:
                print_item(family)
        else:

            for style in family.styles:
                start_check()
                is_match = match_item(style)
                end_check()

                if is_match:
                    print_item(style)



#############################################################################
##
## Copyright (C) 2010 Riverbank Computing Limited.
## Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
## All rights reserved.
##
## This file is part of the examples of PyQt.
##
## $QT_BEGIN_LICENSE:BSD$
## You may use this file under the terms of the BSD license as follows:
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are
## met:
##   * Redistributions of source code must retain the above copyright
##     notice, this list of conditions and the following disclaimer.
##   * Redistributions in binary form must reproduce the above copyright
##     notice, this list of conditions and the following disclaimer in
##     the documentation and/or other materials provided with the
##     distribution.
##   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
##     the names of its contributors may be used to endorse or promote
##     products derived from this software without specific prior written
##     permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
## $QT_END_LICENSE$
##
#############################################################################


from typeatlas.compat import QtCore, QtGui, QtWidgets


class FlowLayout(QtWidgets.QLayout):
    def __init__(self, parent=None, margin=0, spacing=-1):
        super(FlowLayout, self).__init__(parent)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self.setSpacing(spacing)
        self.margin = margin
        
        # spaces between each item
        self.spaceX = 5
        self.spaceY = 5

        self.itemList = []

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addLayout(self, layout):
        w = QtWidgets.QWidget()
        w.setLayout(layout)
        self.addWidget(w)

    def addItem(self, item):
        self.itemList.append(item)

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList[index]

        return None

    def takeAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList.pop(index)

        return None

    def expandingDirections(self):
        return QtCore.Qt.Orientations(QtCore.Qt.Orientation(0))

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self.doLayout(QtCore.QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QtCore.QSize()

        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())

        size += QtCore.QSize(2 * self.margin, 2 * self.margin)
        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0

        for item in self.itemList:
            wid = item.widget()
            # spaceX = self.spacing() + wid.style().layoutSpacing(QtGui.QSizePolicy.PushButton, QtGui.QSizePolicy.PushButton, QtCore.Qt.Horizontal)
            # spaceY = self.spacing() + wid.style().layoutSpacing(QtGui.QSizePolicy.PushButton, QtGui.QSizePolicy.PushButton, QtCore.Qt.Vertical)
            nextX = x + item.sizeHint().width() + self.spaceX
            if nextX - self.spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + self.spaceY
                nextX = x + item.sizeHint().width() + self.spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())

        return y + lineHeight - rect.y()


class FlowLayoutSizePolicy(QtWidgets.QSizePolicy):

    def __init__(self, layout):
        super().__init__(self.MinimumExpanding, self.MinimumExpanding)
        self.layout = layout

    def hasHeightForWidth(self):
        return self.layout.hasHeightForWidth()

    def heightForWidth(self, width):
        return self.layout.heightForWidth(width)

    def hasWidthForHeight(self):
        return self.layout.hasWidthForHeight()

    def hasWidthForHeight(self, height):
        return self.layout.hasWidthForHeight(height)


class HorizontalFlowLayout(FlowLayout):

    minimumHeight = 50
    lastHeight = None

    def sizeHint(self):
        minimumHeight = self.minimumHeight

        if self.lastHeight:
            height = max(minimumHeight, self.lastHeight)
        else:
            height = minimumHeight

        for item in self.itemList:
            minimumHeight = max(minimumHeight, item.minimumSize().height())

        size = QtCore.QSize(self.widthForHeight(height) + 2 * self.margin,
                            minimumHeight + 2 * self.margin)
        return size

    def expandingDirections(self):
        return QtCore.Qt.Orientations(QtCore.Qt.Horizontal)

    def hasHeightForWidth(self):
        return False

    def heightForWidth(self, width):
        return -1

    def hasWidthForHeight(self):
        return True

    def widthForHeight(self, height):
        width = self.doLayout(QtCore.QRect(0, 0, 0, height), True)
        return width

    def doLayout(self, rect, testOnly):
        if not testOnly and self.lastHeight != rect.height():
            self.lastHeight = rect.height()
            self.invalidate()

        x = rect.x()
        y = rect.y()
        lineWidth = 0

        for item in self.itemList:
            wid = item.widget()
            # Reverse of above
            nextY = y + item.sizeHint().height() + self.spaceY
            if nextY - self.spaceY > rect.bottom() and lineWidth > 0:
                y = rect.y()
                x = x + lineWidth + self.spaceX
                nextY = y + item.sizeHint().height() + self.spaceY
                lineWidth = 0

            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))

            y = nextY
            lineWidth= max(lineWidth, item.sizeHint().width())

        return x + lineWidth - rect.x()


